# pylint: disable=invalid-name,too-many-instance-attributes,too-many-locals
import numpy as np
import scipy.special

from ..hasher import VideoHasher, ImageHasher
from ..image.phash import PHashF


class TMKL2(VideoHasher):
    """The TMK L2 video hashing algorithm."""

    dtype = 'float32'
    distance_metric = 'custom'

    def __init__(self,
                 frame_hasher: ImageHasher = None,
                 frames_per_second: int = 15,
                 normalization: str = 'matrix'):
        T = np.array([2731, 4391, 9767, 14653]).astype('float32')
        m = 32
        if frame_hasher is None:
            frame_hasher = PHashF(
                hash_size=16, exclude_first_term=True, freq_shift=1)
        self.frames_per_second = frames_per_second
        assert frame_hasher.dtype != 'bool', 'This hasher requires real valued hashes.'

        # Beta parameter of the modified Bessel function of the first kind
        self.beta = 32

        # Number of Fourier coefficients per period
        self.m = m

        # The periods with shape (T, )
        self.T = T  # (T)

        # The Fourier coefficients with shape (T, m, 1)
        self.ms = 2 * np.pi * np.arange(0, self.m).astype('float32')  # (m)
        self.ms_normed = (
            self.ms[np.newaxis, ] / self.T.reshape(-1, 1)).reshape(
                len(self.T), self.m, 1)  # (T, m, 1)

        # The weights with shape (T, 2m, 1)
        a = np.array([(scipy.special.iv(0, self.beta) - np.exp(-self.beta)) /
                      (2 * np.sinh(self.beta))] + [
                          scipy.special.iv(i, self.beta) / np.sinh(self.beta)
                          for i in range(1, self.m)
                      ])
        a = a.reshape(1, -1).repeat(repeats=len(self.T), axis=0)
        a = np.sqrt(a)
        self.a = a[..., np.newaxis]

        # The frame-wise hasher
        self.frame_hasher = frame_hasher

        # pylint: disable=unsubscriptable-object
        self.hash_length = self.T.shape[
            0] * 2 * self.m * self.frame_hasher.hash_length

        self.normalization = normalization

    def process_frame(self, frame, frame_index, frame_timestamp, state=None):
        if state is None:
            state = {'features': [], 'timestamps': []}
        state['features'].append(
            self.frame_hasher.compute(frame, hash_format='vector'))
        state['timestamps'].append(frame_timestamp)
        return state

    def hash_from_final_state(self, state):
        timestamps = np.array(state['timestamps'])
        # pylint: disable=unsubscriptable-object
        features = np.array(state['features']).reshape(
            (1, 1, timestamps.shape[0], self.frame_hasher.hash_length))
        x = self.ms_normed * timestamps
        yw1 = np.sin(x) * self.a
        yw2 = np.cos(x) * self.a
        yw = np.concatenate([yw1, yw2], axis=1)[...,
                                                np.newaxis]  # (T, 2m, t, 1)
        y = (yw * features).sum(axis=2)  # (T, 2m, d)
        return y.flatten()

    def _compute_distance(self, vector1, vector2):
        shape = (len(self.T), 2 * self.m, self.frame_hasher.hash_length)
        return 1 - self._score_pair(
            fv_a=vector1.reshape(shape),
            fv_b=vector2.reshape(shape),
            offsets=None,
            normalization=self.normalization)

    def _score_pair(self, fv_a, fv_b, offsets=None, normalization='matrix'):
        eps = 1e-8

        if offsets is None:
            offsets = np.array([0])

        assert normalization in ['feat', 'freq', 'feat_freq',
                                 'matrix'], 'Invalid normalization'

        if "feat" in normalization:
            a_xp = np.concatenate([self.a, self.a], axis=1)  # (T, 2m, 1)
            fv_a_0 = fv_a / a_xp
            fv_b_0 = fv_b / a_xp
            norm_a = np.sqrt(np.sum(fv_a_0**2, axis=2, keepdims=True) +
                             eps) + eps
            norm_b = np.sqrt(np.sum(fv_b_0**2, axis=2, keepdims=True) +
                             eps) + eps
            fv_a = fv_a / norm_a
            fv_b = fv_b / norm_b

        if "freq" in normalization:
            norm_a, norm_b = [
                np.sqrt((fv**2).sum(axis=1, keepdims=True) / self.m + eps) +
                eps for fv in [fv_a, fv_b]
            ]
            fv_a = fv_a / norm_a
            fv_b = fv_b / norm_b

        if normalization == "matrix":
            norm_a, norm_b = [
                np.sqrt(np.sum(fv**2, axis=(1, 2)) + eps)[..., np.newaxis] +
                eps for fv in [fv_a, fv_b]
            ]  # (T, 1)

        fv_a_sin, fv_b_sin = [fv[:, :self.m]
                              for fv in [fv_a, fv_b]]  # (T, m, d)
        fv_a_cos, fv_b_cos = [fv[:, self.m:]
                              for fv in [fv_a, fv_b]]  # (T, m, d)
        ms = self.ms.reshape(-1, 1)  # (m, 1)
        dot_sin_sin, dot_sin_cos, dot_cos_cos, dot_cos_sin = [
            np.sum(p, axis=2, keepdims=True) for p in [
                fv_a_sin * fv_b_sin, fv_a_sin * fv_b_cos, fv_a_cos *
                fv_b_cos, fv_a_cos * fv_b_sin
            ]
        ]  # (T, m, 1)
        delta = ms.reshape(1, -1, 1) * offsets.reshape(1, -1) / self.T.reshape(
            (-1, 1, 1))
        cos_delta = np.cos(delta)  # (T, m, delta)
        sin_delta = np.sin(delta)  # (T, m, delta)
        dots = (dot_sin_sin * cos_delta + dot_sin_cos * sin_delta +
                dot_cos_cos * cos_delta - dot_cos_sin * sin_delta).sum(axis=1)
        if normalization == "matrix":
            dots = dots / (norm_a * norm_b)
        if normalization == "freq":
            dots = dots / self.m  # (T, m, delta)
        elif normalization in ["feat", "feat_freq"]:
            dots = dots / 512
        return dots.mean(axis=0)


class TMKL1(VideoHasher):
    """The TMK L1 video hashing algorithm."""

    # pylint: disable=too-many-arguments
    def __init__(self,
                 frame_hasher: ImageHasher = None,
                 frames_per_second: int = 15,
                 dtype='float32',
                 distance_metric='cosine',
                 norm=2,
                 quality_threshold=None):
        if frame_hasher is None:
            frame_hasher = PHashF(
                hash_size=16, exclude_first_term=True, freq_shift=1)
        self.hash_length = frame_hasher.hash_length
        self.frames_per_second = frames_per_second
        assert frame_hasher.dtype != 'bool', 'This hasher requires real valued hashes.'
        self.frame_hasher = frame_hasher
        self.norm = norm
        self.dtype = dtype or self.frame_hasher.dtype
        self.distance_metric = distance_metric or self.frame_hasher.distance_metric
        self.quality_threshold = quality_threshold

    # pylint: disable=unused-argument
    def process_frame(self, frame, frame_index, frame_timestamp, state=None):
        if state is None:
            state = {
                'sum': np.zeros(self.frame_hasher.hash_length),
                'frame_count': 0
            }
        if not self.quality_threshold:
            hash_vector = self.frame_hasher.compute(
                frame, hash_format='vector')
        else:
            hash_vector, quality = self.frame_hasher.compute_with_quality(
                frame, hash_format='vector')
            if quality < self.quality_threshold:
                return state
        if hash_vector is not None:
            state['sum'] += np.float32(hash_vector)
            state['frame_count'] += 1
        return state

    def hash_from_final_state(self, state):
        if state['frame_count'] == 0:
            return None
        average_vector = state['sum'] / state['frame_count']
        if self.norm is not None:
            return (average_vector / np.linalg.norm(
                average_vector, ord=self.norm)).astype(self.frame_hasher.dtype)
        return average_vector.astype(self.frame_hasher.dtype)



import typing
import warnings
import logging
import uuid
import os

from tqdm import tqdm
import pandas as pd
import imgaug
import cv2

from ..hashers import ImageHasher, tools
from ..tools import deduplicate, flatten
from .common import BenchmarkTransforms, BenchmarkDataset, BenchmarkHashes

# pylint: disable=invalid-name
log = logging.getLogger(__name__)


class BenchmarkImageTransforms(BenchmarkTransforms):
    def compute_hashes(self,
                       hashers: typing.Dict[str, ImageHasher],
                       max_workers: int = 5) -> BenchmarkHashes:
        """Compute hashes for a series of files given some set of hashers.

        Args:
            hashers: A dictionary of hashers.
            max_workers: Maximum number of workers for parallel hash
                computation.

        Returns:
            metrics: A BenchmarkHashes object.
        """
        hashsets = []
        filepaths = self._df['filepath']
        for hasher_name, hasher in hashers.items():
            hash_dicts = hasher.compute_parallel(
                filepaths,
                progress=tqdm,
                progress_desc=f'Computing hashes for {hasher_name}',
                max_workers=max_workers)
            if not hasher.returns_multiple:
                hashes_df = pd.DataFrame.from_records(hash_dicts)
            else:
                hash_groups = [
                    hash_dict['hash']
                    if hash_dict['error'] is None else [None]
                    for hash_dict in hash_dicts
                ]
                hash_group_sizes = [
                    len(hash_group) for hash_group in hash_groups
                ]
                current_hashes = flatten(hash_groups)
                current_filepaths = flatten(
                    [[hash_dict['filepath']] * hash_group_size
                     for hash_dict, hash_group_size in zip(
                         hash_dicts, hash_group_sizes)])
                current_errors = flatten(
                    [[hash_dict['error']] * hash_group_size
                     for hash_dict, hash_group_size in zip(
                         hash_dicts, hash_group_sizes)])
                hashes_df = pd.DataFrame({
                    'error': current_errors,
                    'filepath': current_filepaths,
                    'hash': current_hashes
                })
            hashset = hashes_df.assign(
                hasher_name=hasher_name,
                hasher_hash_length=hasher.hash_length,
                hasher_dtype=hasher.dtype,
                hasher_distance_metric=hasher.distance_metric)
            hashset = hashset.merge(self._df, on='filepath')
            hashsets.append(hashset)
        return BenchmarkHashes(pd.concat(hashsets, sort=True))


class BenchmarkImageDataset(BenchmarkDataset):
    # pylint: disable=too-many-locals
    def deduplicate(self,
                    hasher: ImageHasher,
                    threshold=0.001,
                    isometric=False
                    ) -> typing.Tuple['BenchmarkImageDataset', typing.
                                      Set[typing.Tuple[str, str]]]:
        """ Remove duplicate files from dataset.

        Args:
            files: A list of file paths
            hasher: A hasher to use for finding a duplicate
            threshold: The threshold required for a match
            isometric: Whether to compute the rotated versions of the images

        Returns:
            A list where each entry is a list of files that are
            duplicates of each other. We keep only the last entry.
        """
        pairs: typing.Set[typing.Tuple[str, str]] = set()
        for _, group in tqdm(
                self._df.groupby(['category']),
                desc='Deduplicating categories.'):
            pairs = pairs.union(
                set(
                    deduplicate(
                        files=group['filepath'],
                        hashers=[(hasher, threshold)],
                        isometric=isometric)))
        removed = [pair[0] for pair in pairs]
        return BenchmarkImageDataset(
            self._df[~self._df['filepath'].isin(removed)].copy()), pairs

    def transform(
            self,
            transforms: typing.Dict[str, imgaug.augmenters.meta.Augmenter],
            storage_dir: str,
            errors: str = "raise") -> BenchmarkImageTransforms:
        """Prepare files to be used as part of benchmarking run.

        Args:
            transforms: A dictionary of transformations. The only required
                key is `noop` which determines how the original, untransformed
                image is saved. For a true copy, simply make the `noop` key
                `imgaug.augmenters.Noop()`.
            storage_dir: A directory to store all the images along with
                their transformed counterparts.
            errors: How to handle errors reading files. If "raise", exceptions are
                raised. If "warn", the error is printed as a warning.

        Returns:
            transforms: A BenchmarkImageTransforms object
        """
        assert 'noop' in transforms, 'You must provide a no-op transform such as `lambda img: img`.'

        os.makedirs(storage_dir, exist_ok=True)

        files = self._df.copy()
        files['guid'] = [uuid.uuid4() for n in range(len(files))]

        def apply_transform(files, transform_name):
            transform = transforms[transform_name]
            transformed_arr = []
            for _, row in tqdm(
                    files.iterrows(),
                    desc=f'Creating files for {transform_name}',
                    total=len(files)):
                filepath, guid, category = row[[
                    'filepath', 'guid', 'category'
                ]]
                try:
                    image = tools.read(filepath)
                # pylint: disable=broad-except
                except Exception as exception:
                    message = f'An error occurred reading {filepath}.'
                    if errors == 'raise':
                        raise exception
                    warnings.warn(message, UserWarning)
                    continue
                try:
                    transformed = transform(image=image)
                except Exception as e:
                    raise Exception(
                        f'An exception occurred while processing {filepath} '
                        f'with transform {transform_name}.') from e
                transformed_path = os.path.join(
                    storage_dir, f'{guid}_{transform_name}.jpg')
                cv2.imwrite(transformed_path,
                            cv2.cvtColor(transformed, cv2.COLOR_RGB2BGR))
                transformed_arr.append({
                    'guid': guid,
                    'transform_name': transform_name,
                    'input_filepath': filepath,
                    'filepath': transformed_path,
                    'category': category
                })
            return pd.DataFrame.from_records(transformed_arr)

        results = [apply_transform(files, transform_name='noop')]

        for transform_name in transforms.keys():
            if transform_name == 'noop':
                continue
            results.append(
                apply_transform(results[0], transform_name=transform_name))
        benchmark_transforms = BenchmarkImageTransforms(
            df=pd.concat(results, axis=0, ignore_index=True))
        benchmark_transforms.save(storage_dir)
        return benchmark_transforms



# pylint: disable=invalid-name,too-many-locals
import os
import math
import typing
import pkg_resources

import cv2
import pytest
import numpy as np
import pandas as pd
from PIL import Image  # pylint: disable=import-error

from .. import hashers, tools

SIZES = {'float32': 32, 'uint8': 8, 'bool': 1}


def get_low_detail_image():
    v = np.arange(0, 50, 1)
    v = np.concatenate([v, v[::-1]])[np.newaxis, ]
    image = np.matmul(v.T, v)
    image = (image * 255 / image.max()).astype('uint8')
    image = image[..., np.newaxis].repeat(repeats=3, axis=2)
    image[:, 50:] = 0
    image[50:] = 0
    return image


LOW_DETAIL_IMAGE = get_low_detail_image()

DEFAULT_TEST_IMAGES = [
    pkg_resources.resource_filename(
        'perception', os.path.join('testing', 'images', f'image{n}.jpg'))
    for n in range(1, 11)
]
DEFAULT_TEST_VIDEOS = [
    pkg_resources.resource_filename(
        'perception', os.path.join('testing', 'videos', f'v{n}.m4v'))
    for n in range(1, 3)
] + [
    pkg_resources.resource_filename(
        'perception', os.path.join('testing', 'videos', 'v2s.mov'))
]


@typing.no_type_check
def test_opencv_hasher(hasher: hashers.ImageHasher, image1: str, image2: str):
    # For OpenCV hashers we make sure the distance we compute
    # is the same as inside OpenCV
    f1 = image1
    f2 = image2
    opencv_distance = hasher.hasher.compare(
        hasher.hasher.compute(hashers.tools.read(f1)),
        hasher.hasher.compute(hashers.tools.read(f2)))
    if hasher.distance_metric == 'hamming':
        opencv_distance /= hasher.hash_length
    np.testing.assert_approx_equal(
        opencv_distance,
        hasher.compute_distance(hasher.compute(f1), hasher.compute(f2)),
        significant=4)


# pylint: disable=protected-access
def hash_dicts_to_df(hash_dicts, returns_multiple):
    assert all(
        h['error'] is None
        for h in hash_dicts), 'An error was found in the hash dictionaries'
    if returns_multiple:
        return pd.DataFrame({
            'filepath':
            tools.flatten(
                [[h['filepath']] * len(h['hash']) for h in hash_dicts]),
            'hash':
            tools.flatten([h['hash'] for h in hash_dicts])
        }).assign(error=None)
    return pd.DataFrame.from_records(hash_dicts).assign(error=None)


def test_hasher_parallelization(hasher, test_filepaths):
    filepaths_10x = test_filepaths * 10
    if not hasher.allow_parallel:
        with pytest.warns(UserWarning, match='cannot be used in parallel'):
            hashes_parallel_dicts = hasher.compute_parallel(
                filepaths=filepaths_10x)
    else:
        hashes_parallel_dicts = hasher.compute_parallel(
            filepaths=filepaths_10x)
    hashes_sequential_dicts = [{
        'filepath': filepath,
        'hash': hasher.compute(filepath),
        'error': None
    } for filepath in filepaths_10x]
    hashes_parallel = hash_dicts_to_df(
        hashes_parallel_dicts,
        returns_multiple=hasher.returns_multiple).sort_values(
            ['filepath', 'hash'])
    hashes_sequential = hash_dicts_to_df(
        hashes_sequential_dicts,
        returns_multiple=hasher.returns_multiple).sort_values(
            ['filepath', 'hash'])
    assert (hashes_sequential.hash.values == hashes_parallel.hash.values).all()
    assert (hashes_sequential.filepath.values == hashes_parallel.filepath.
            values).all()


def test_video_hasher_integrity(hasher: hashers.VideoHasher,
                                test_videos: typing.List[str] = None):
    if test_videos is None:
        test_videos = DEFAULT_TEST_VIDEOS
    test_hasher_parallelization(hasher, test_videos)


def test_image_hasher_integrity(hasher: hashers.ImageHasher,
                                pil_opencv_threshold: float,
                                transform_threshold: float,
                                test_images: typing.List[str] = None,
                                opencv_hasher: bool = False):
    """Test to ensure a hasher works correctly.

    Args:
        hasher: The hasher to test.
        test_images: A list of filepaths to images to use for testing.
        pil_opencv_threshold: The hash distance permitted for an image
            when loaded with OpenCV vs. PIL.
        transform_threshold: The permitted error in isometric transform
            hashes.
        opencv_hasher: Whether the hasher is an OpenCV hasher. Used to
            determine whether to check for consistent distances.
    """
    if test_images is None:
        test_images = DEFAULT_TEST_IMAGES
    assert len(test_images) >= 2, 'You must provide at least two test images.'
    image1 = test_images[0]
    image2 = test_images[1]
    hash1_1 = hasher.compute(image1)
    hash1_2 = hasher.compute(Image.open(image1))
    hash1_3 = hasher.compute(
        cv2.cvtColor(cv2.imread(image1), cv2.COLOR_BGR2RGB))

    hash2_1 = hasher.compute(image2)

    # There is a small distance because PIL and OpenCV read
    # JPEG images a little differently (e.g., libjpeg-turbo vs. libjpeg)
    assert hasher.compute_distance(hash1_1, hash1_2) < pil_opencv_threshold
    assert hasher.compute_distance(hash1_1, hash2_1) > pil_opencv_threshold
    assert hasher.compute_distance(hash1_1, hash1_3) == 0

    # Ensure the conversion to and from vectors works for both base64 and hex.
    assert hasher.vector_to_string(hasher.string_to_vector(hash2_1)) == hash2_1
    assert hasher.vector_to_string(
        hasher.string_to_vector(
            hasher.vector_to_string(
                hasher.string_to_vector(hash2_1), hash_format='hex'),
            hash_format='hex')) == hash2_1

    # Ensure parallelization works properly.
    test_hasher_parallelization(hasher=hasher, test_filepaths=test_images)

    # Ensure the isometric hashes computation work properly
    for image in test_images:
        transforms = hashers.tools.get_isometric_transforms(image)
        hashes_exp = {
            key: hasher.compute(value)
            for key, value in transforms.items()
        }
        hashes_act = hasher.compute_isometric(transforms['r0'])
        for transform_name in hashes_exp.keys():
            assert hasher.compute_distance(
                hashes_exp[transform_name],
                hashes_act[transform_name]) < transform_threshold

    # Verify that hashes are the correct length.
    hash_bits = hasher.hash_length * SIZES[hasher.dtype]

    words_base64 = math.ceil(
        hash_bits / 6)  # Base64 uses 8 bits for every 6 bits
    words_base64 += 0 if words_base64 % 4 == 0 else 4 - (
        words_base64 % 4)  # Base64 always uses multiples of four
    assert len(hash2_1) == words_base64

    words_hex = 2 * math.ceil(
        hash_bits / 8)  # Hex uses 16 bits for every 8 bits
    words_hex += 0 if words_hex % 2 == 0 else 1  # Two characters for every one character.
    assert len(
        hasher.vector_to_string(
            hasher.string_to_vector(hash2_1), hash_format='hex')) == words_hex

    # Verify that low quality images yield zero quality
    image = np.zeros((100, 100, 3)).astype('uint8')
    _, quality = hasher.compute_with_quality(image)
    assert quality == 0

    # Verify that high quality images yield high quality
    # scores.
    assert min(
        hasher.compute_with_quality(filepath)[1]
        for filepath in test_images) == 100

    # Verify that medium quality images yield medium quality
    _, quality = hasher.compute_with_quality(LOW_DETAIL_IMAGE)
    assert 0 < quality < 100

    if opencv_hasher:
        test_opencv_hasher(hasher, image1, image2)



# pylint: disable=no-member
from abc import ABC, abstractmethod
import typing
import concurrent.futures
import warnings

import numpy as np
import scipy.spatial
try:
    import tqdm  # pylint: disable=unused-import
except ImportError:  # pragma: no cover
    tqdm = None
try:
    # We do this so that the documentation generator can
    # resolve the forward reference to PIL in development.
    import PIL  # pylint: disable=unused-import
    import PIL.Image  # pylint: disable=unused-import
except ImportError:  # pragma: no cover
    PIL = None

from perception.hashers import tools


class Hasher(ABC):
    """All hashers implement a common set of methods from
    the Hasher base class.
    """

    #: The metric to use when computing distance between two hashes. All hashers
    #: must supply this parameter.
    distance_metric: str

    #: The numpy type to use when converting from string to array form.
    #: All hashers must supply this parameter.
    dtype: str

    #: Indicates the length of the hash vector
    hash_length: int

    #: Whether or not this hash returns multiple values
    returns_multiple: bool = False

    #: Indicates whether the hashes can be computed in parallel
    allow_parallel: bool = True

    def string_to_vector(self, hash_string: str, hash_format: str = 'base64'):
        """Convert hash string to vector.

        Args:
            hash_string: The input hash string
            hash_format: One of 'base64' or 'hex'
        """
        return tools.string_to_vector(
            hash_string,
            dtype=self.dtype,
            hash_length=self.hash_length,
            hash_format=hash_format)

    def vector_to_string(self, vector: np.ndarray,
                         hash_format: str = 'base64'):
        """Convert vector to hash string.

        Args:
            vector: Input vector
            hash_format: One of 'base64' or 'hex'
        """
        return tools.vector_to_string(
            vector, dtype=self.dtype, hash_format=hash_format)

    def compute_distance(self,
                         hash1: typing.Union[np.ndarray, str],
                         hash2: typing.Union[np.ndarray, str],
                         hash_format='base64'):
        """Compute the distance between two hashes.

        Args:
            hash1: The first hash or vector
            hash2: The second hash or vector
            hash_format: If either or both of the hashes are hash strings,
                what format the string is encoded in.
        """
        if isinstance(hash1, str):
            hash1 = self.string_to_vector(hash1, hash_format=hash_format)
        if isinstance(hash2, str):
            hash2 = self.string_to_vector(hash2, hash_format=hash_format)

        if self.distance_metric == 'sqeuclidean':
            return scipy.spatial.distance.sqeuclidean(
                hash1.astype('float32'), hash2.astype('float32'))
        if self.distance_metric == 'euclidean':
            return scipy.spatial.distance.euclidean(
                hash1.astype('float32'), hash2.astype('float32'))
        if self.distance_metric == 'hamming':
            return scipy.spatial.distance.hamming(hash1, hash2)
        if self.distance_metric == 'cosine':
            return scipy.spatial.distance.cosine(
                hash1.astype('float32'), hash2.astype('float32'))
        if self.distance_metric == 'custom':
            return self._compute_distance(hash1, hash2)
        raise NotImplementedError(
            f'Distance metric: {self.distance_metric} not supported.')

    # pylint: disable=no-self-use
    def _compute_distance(self, vector1, vector2):
        raise ValueError(
            'Called a custom distance function but it is not implemented.')

    # pylint: disable=too-many-arguments,too-many-locals
    @typing.no_type_check
    def compute_parallel(self,
                         filepaths: typing.List[str],
                         progress: 'tqdm.tqdm' = None,
                         progress_desc: str = None,
                         max_workers: int = 5,
                         isometric: bool = False):
        """Compute hashes in a parallelized fashion.

        Args:
            filepaths: A list of paths to images or videos (depending on the hasher).
            progress: A tqdm-like wrapper for reporting progress. If None,
                progress is not reported.
            progress_desc: The title of the progress bar.
            max_workers: The maximum number of workers
            isometric: Whether to compute all eight isometric transforms for
                each image.
        """
        if not self.allow_parallel and max_workers != 1:
            warnings.warn(
                message=
                'This hash cannot be used in parallel. Setting max_workers to 1.',
                category=UserWarning)
            max_workers = 1
        assert all(
            isinstance(p, str)
            for p in filepaths), 'All images should be provided as paths.'

        if isinstance(self, VideoHasher) and isometric:
            raise ValueError(
                'Computing isometric hashes for videos is not supported.')

        # We can use a with statement to ensure threads are cleaned up promptly
        records = []
        if isinstance(self, VideoHasher):
            executor_class = concurrent.futures.ProcessPoolExecutor
        else:
            executor_class = concurrent.futures.ThreadPoolExecutor
        with executor_class(max_workers=max_workers) as executor:
            # Start the load operations and mark each future with its filepath
            compute: typing.Callable = self.compute_isometric if isometric else self.compute
            future_to_path: dict = {
                executor.submit(compute, path): path
                for path in filepaths
            }
            generator = concurrent.futures.as_completed(future_to_path)
            if progress is not None:
                generator = progress(
                    generator, total=len(filepaths), desc=progress_desc)
            for future in generator:
                path = future_to_path[future]
                try:
                    hash_value = future.result()
                # pylint: disable=broad-except
                except Exception as exc:
                    records.append({
                        'filepath': path,
                        'hash': None,
                        'error': str(exc)
                    })
                else:
                    records.append({
                        'filepath': path,
                        'hash': hash_value,
                        'error': None
                    })
        return records


class ImageHasher(Hasher):
    @abstractmethod
    def _compute(self, image: np.ndarray):
        """Compute hash from an image.

        Args:
            image: A numpy array representing an image as
                of shape (H, W, 3) where channels are ordered
                as RGB or a filepath to an image.
        """

    def compute_isometric_from_hash(self,
                                    hash_string_or_vector,
                                    hash_format='base64'):
        """For supported hashes, obtain the hashes for the dihedral transformations
        of the original image. They are provided in the following order:

        - Vertical flip
        - Horizontal flip
        - 180 degree rotation
        - 90 degree rotation
        - 90 degree rotation and vertical flip
        - 90 degree rotation and horizontal flip
        - 270 degree rotation

        Args:
            hash_string_or_vector: The hash string or vector
            hash_format: One 'base64' or 'hex'
        """
        if not hasattr(self, '_compute_isometric_from_hash'):
            raise NotImplementedError(
                "This hasher does not support hash rotation.")
        # pylint: disable=no-member
        rotations = self._compute_isometric_from_hash(  # type: ignore
            hash_string_or_vector if isinstance(
                hash_string_or_vector, np.ndarray) else self.string_to_vector(
                    hash_string_or_vector, hash_format=hash_format))
        return {
            transform_name: self.vector_to_string(
                vector, hash_format=hash_format)
            for transform_name, vector in rotations.items()
        }

    def compute_isometric(self, image: tools.ImageInputType):
        image = tools.to_image_array(image)
        if hasattr(self, '_compute_isometric'):
            # pylint: disable=no-member
            hashes = self._compute_isometric(image)  # type: ignore
        elif hasattr(self, '_compute_isometric_from_hash'):
            # pylint: disable=no-member
            hashes = self._compute_isometric_from_hash(  # type: ignore
                self._compute(image))
        else:
            transforms = tools.get_isometric_transforms(image)
            for name, transform in transforms.items():
                transforms[name] = self._compute(transform)
            hashes = transforms
        return {
            transform_name: self.vector_to_string(vector)
            for transform_name, vector in hashes.items()
        }

    def compute(self, image: tools.ImageInputType,
                hash_format='base64') -> typing.Union[str, np.ndarray]:
        """Compute a hash from an image.

        Args:
            image: An image represented as a filepath, a PIL image object,
                or as an np.ndarray object. If it is an np.ndarray object,
                it must be in RGB color order (note the OpenCV default is
                BGR).
            hash_format: One 'base64', 'hex', or 'vector'
        """
        vector = self._compute(tools.to_image_array(image))
        return self.vector_to_string(
            vector,
            hash_format=hash_format) if not self.returns_multiple else [
                self.vector_to_string(v, hash_format=hash_format)
                for v in vector
            ]

    def compute_with_quality(self,
                             image: tools.ImageInputType,
                             hash_format='base64') -> typing.Tuple[str, int]:
        """Compute hash and hash quality from image.

        Args:
            image: An image represented as a filepath, a PIL image object,
                or as an np.ndarray object. If it is an np.ndarray object,
                it must be in RGB color order (note the OpenCV default is
                BGR).
            hash_format: One 'base64' or 'hex'

        Returns:
            A tuple of (hash, quality)
        """
        vector, quality = self._compute_with_quality(
            tools.to_image_array(image))
        if hash_format == 'vector':
            return vector, quality
        return (self.vector_to_string(vector, hash_format=hash_format),
                quality) if not self.returns_multiple else ([
                    self.vector_to_string(v, hash_format=hash_format)
                    for v in vector
                ], quality)

    def _compute_with_quality(self, image: np.ndarray):
        return self._compute(image), tools.compute_quality(image)


class VideoHasher(Hasher):

    #: The frame rate at which videos are read
    frames_per_second: float = 1

    @abstractmethod
    def process_frame(self,
                      frame: np.ndarray,
                      frame_index: int,
                      frame_timestamp: float,
                      state: dict = None) -> dict:
        """Called for each frame in the video. For all
        but the first frame, a state is provided recording the state from
        the previous frame.

        Args:
            frame: The current frame as an RGB ndarray
            frame_index: The current frame index
            frame_timestamp: The current frame timestamp
            state: The state from the last call to process_frame
        """

    @abstractmethod
    def hash_from_final_state(self, state: dict) -> np.ndarray:
        """Called after all frames have been processed. Returns the final
        feature vector.

        Args:
            state: The state dictionary at the end of processing.
        """

    # pylint: disable=arguments-differ
    def compute(self, filepath, errors='raise', hash_format='base64',
                **kwargs):
        """Compute a hash for a video at a given filepath. All
        other arguments are passed to perception.hashers.tools.read_video.

        Args:
            filepath: Path to video file
            errors: One of "raise", "ignore", or "warn". Passed
                to perception.hashers.tools.read_video.
            hash_format: One of "vector", "base64", or "hex"
            max_duration: The maximum length of the video to hash.
            max_size: The maximum size of frames to queue
        """
        state = None
        for frame, frame_index, frame_timestamp in tools.read_video(
                filepath=filepath,
                frames_per_second=self.frames_per_second,
                errors=errors,
                **kwargs):
            state = self.process_frame(
                frame=frame,
                frame_index=frame_index,
                frame_timestamp=frame_timestamp,
                state=state)
        assert state is not None
        vector = self.hash_from_final_state(state=state)
        return self.vector_to_string(
            vector,
            hash_format=hash_format) if not self.returns_multiple else [
                self.vector_to_string(v, hash_format=hash_format)
                for v in vector
            ]



from .hasher import Hasher, ImageHasher, VideoHasher
from .image.average import AverageHash
from .image.phash import PHash, PHashU8, PHashF
from .image.wavelet import WaveletHash
from .image.opencv import MarrHildreth, BlockMean, ColorMoment
from .image.dhash import DHash
from .video.framewise import FramewiseHasher
from .video.scenes import SimpleSceneDetection
from .video.tmk import TMKL1, TMKL2

__all__ = [
    'AverageHash', 'PHash', 'WaveletHash', 'MarrHildreth', 'BlockMean',
    'ColorMoment', 'DHash', 'FramewiseHasher', 'TMKL1', 'TMKL2', 'PHashU8',
    'PHashF', 'SimpleSceneDetection'
]



import cv2

from ..hasher import ImageHasher


class DHash(ImageHasher):
    """A hash based on the differences between adjacent pixels.
    Implementation based on that of
    `ImageHash <https://github.com/JohannesBuchner/imagehash>`_.
    """
    dtype = 'bool'
    distance_metric = 'hamming'

    def __init__(self, hash_size=8):
        assert hash_size > 1, 'Hash size must be greater than 1.'
        self.hash_size = hash_size
        self.hash_length = hash_size * hash_size

    def _compute(self, image):
        image = cv2.resize(
            image,
            dsize=(self.hash_size + 1, self.hash_size),
            interpolation=cv2.INTER_AREA)
        image = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
        previous = image[:, :-1]
        current = image[:, 1:]
        difference = previous > current
        return difference.flatten()



# distutils: extra_compile_args=-fopenmp
# distutils: extra_link_args=-fopenmp
# cython: language_level=3

import numpy as np
import cython
from cython.parallel import prange, parallel
from libc.stdlib cimport abort, malloc, free
from libc.math cimport sqrt

cimport numpy as np
cdef extern from "limits.h":
    int INT_MAX

ctypedef np.uint8_t uint8

@cython.boundscheck(False)
@cython.wraparound(False)
def compute_euclidean_metrics(int[:, :] X_noop, int[:, :] X_tran, uint8[:, :] mask):
    """Compute the positive / negative distance metrics between two sets of vectors
    using euclidean distance. This function obtains the necessary metrics roughly
    10x faster than using scipy.spatial.distance.cdist and numpy functions.
    
    Args:
        X_noop: The vectors for the noop hashes with shape (N, K)
        X_tran: The vectors for the transformed instances with shape (M, K)
        mask: A (M, N) array indicating whether noop n corresponds to transform m
    
    Returns:
        distances: An M by 2 array with the closest false positive and closest
            true positive for each transform.
        indexes: An M by 2 array with the index for the closest false positive
            noop and the closest true positive noop.
    """

    cdef Py_ssize_t n_noop = X_noop.shape[0]
    cdef Py_ssize_t d_noop = X_noop.shape[1]
    cdef Py_ssize_t n_tran = X_tran.shape[0]
    cdef Py_ssize_t d_tran = X_tran.shape[1]
    cdef Py_ssize_t n_mask_tran = mask.shape[0]
    cdef Py_ssize_t n_mask_noop = mask.shape[1]
    cdef Py_ssize_t i_mask_tran
    cdef Py_ssize_t i_mask_noop
    cdef int n_pos

    cdef int current_distance
    cdef int current_closest_fp
    cdef int current_closest_tp
    cdef int[:] x
    cdef int[:] y
    cdef uint8 is_pos
    cdef Py_ssize_t i_noop, i_tran, i_d
    cdef Py_ssize_t i_closest_fp = 0
    cdef Py_ssize_t i_closest_tp = 1
    cdef Py_ssize_t i_closest_fp_idx = 0
    cdef Py_ssize_t i_closest_tp_idx = 1
    cdef int * local_buf
    cdef size_t size = 5
    cdef float NAN
    NAN = float("NaN")

    assert d_noop == d_tran, "Dimensionality of vectors must match."
    assert n_mask_tran == n_tran, "Dimension 0 of mask must correspond to n_transforms."
    assert n_mask_noop == n_noop, "Dimension 1 of mask must correspond to n_noops."
    for i_mask_tran in range(n_mask_tran):
        n_pos = 0
        for i_mask_noop in range(n_mask_noop):
            if mask[i_mask_tran, i_mask_noop] == True:
                n_pos += 1
        assert n_pos > 0, "All transforms must have at least one positive noop."
        assert n_pos < n_mask_noop, "All transforms must have at least one negative noop."
    
    distances = np.zeros((n_tran, 2), dtype=np.float32)
    indexes = np.zeros((n_tran, 2), dtype=np.int32)
    
    cdef np.float32_t[:, :] distances_view = distances
    cdef int[:, :] indexes_view = indexes

    with nogil, parallel():
        local_buf = <int *> malloc(sizeof(int) * size)
        if local_buf is NULL:
            abort()
        for i_tran in prange(n_tran):
            local_buf[1] = INT_MAX  # Smallest false positive distance
            local_buf[2] = INT_MAX  # Smallest true positive distance
            local_buf[3] = 0        # Smallest false positive index
            local_buf[4] = 0        # Smallest true positive index
            for i_noop in range(n_noop):
                local_buf[0] = 0    # Current distance
                is_pos = mask[i_tran, i_noop] == True
                for i_d in range(d_noop):
                    local_buf[0] += (X_noop[i_noop, i_d] - X_tran[i_tran, i_d]) ** 2
                if is_pos and (local_buf[0] < local_buf[2]):
                    local_buf[2] = local_buf[0]
                    local_buf[4] = i_noop
                if not is_pos and (local_buf[0] < local_buf[1]):
                    local_buf[1] = local_buf[0]
                    local_buf[3] = i_noop
            if local_buf[3] < INT_MAX:
                distances_view[i_tran, i_closest_fp] = sqrt(local_buf[1])
            else:
                distances_view[i_tran, i_closest_fp] = NAN
            if local_buf[2] < INT_MAX:
                distances_view[i_tran, i_closest_tp] = sqrt(local_buf[2])
            else:
                distances_view[i_tran, i_closest_tp] = NAN
            indexes_view[i_tran, i_closest_fp_idx] = local_buf[3]
            indexes_view[i_tran, i_closest_tp_idx] = local_buf[4]
        free(local_buf)
    return distances, indexes


# pylint: disable=too-many-locals
import os
import io
import math
import json
import queue
import base64
import typing
import hashlib
import warnings
import threading
import functools
import itertools
import subprocess
from urllib import request
from http import client

import numpy as np
import validators
import cv2

try:
    import PIL
    import PIL.Image
except ImportError:  # pragma: no cover
    PIL = None

ImageInputType = typing.Union[str, np.ndarray, 'PIL.Image.Image', io.BytesIO]

SIZES = {'float32': 32, 'uint8': 8, 'bool': 1}


# pylint: disable=invalid-name
def compute_quality(image):
    """Compute a quality metric, using the calculation proposed by
    `Facebook <https://github.com/facebook/ThreatExchange/blob/master/hashing/hashing.pdf/>`_
    for their PDQ hash algorithm."""
    if len(image.shape) == 3:
        image = cv2.cvtColor(image, code=cv2.COLOR_RGB2GRAY)
    if image.shape[0] != 64 or image.shape[1] != 64:
        image = cv2.resize(src=image, dsize=(64, 64)).astype('float32')
    dx = 100 * np.abs(image[:, 1:] - image[:, :-1]) / 255
    dy = 100 * np.abs(image[1:] - image[:-1]) / 255
    dx = dx.astype('int').sum()
    dy = dy.astype('int').sum()
    return np.clip(a=int((dx + dy) / 90), a_min=0, a_max=100)


def compute_md5(filepath) -> str:
    """Compute the md5 hash for a file at `filepath`.

    Args:
        filepath: The path to the file
    """
    with open(filepath, 'rb') as f:  # pylint: disable=invalid-name
        hash_str = hashlib.md5(f.read()).hexdigest()
    return hash_str


def get_string_length(hash_length: int, dtype: str, hash_format='hex') -> int:
    """Compute the expected length of a hash string.

    Args:
        hash_length: The length of the hash vector
        dtype: The dtype of the vector
        hash_format: One of 'base64' or 'hex'

    Returns:
        The expected string length
    """
    hash_bytes = math.ceil(hash_length * SIZES[dtype] / 8)

    if hash_format == 'base64':
        return int((4 * hash_bytes / 3) + 3) & ~3
    if hash_format == 'hex':
        return 2 * hash_bytes
    raise NotImplementedError('Unknown hash format: ' + hash_format)


def vector_to_string(vector: np.ndarray, dtype: str, hash_format: str):
    """Convert vector to hash.

    Args:
        vector: Input vector
    """
    # At times, a vector returned by a hasher is None (e.g., for hashes
    # that depend on the image not being featureless). In those cases,
    # we need to just return None, which is the least surprising outcome
    # because after all, the string representation of None is None.
    if vector is None:
        return vector
    if hash_format == 'vector':
        return vector.astype(dtype)
    if dtype == 'uint8':
        vector_bytes = vector.astype('uint8')
    elif dtype == 'float32':
        vector_bytes = vector.astype('float32')
    elif dtype == 'bool':
        vector_bytes = np.packbits(vector.astype('bool'))
    else:
        raise NotImplementedError(f'Cannot convert hash of type {dtype}.')
    if hash_format == 'base64':
        return base64.b64encode(vector_bytes).decode('utf-8')
    if hash_format == 'hex':
        return vector_bytes.tobytes().hex()
    raise NotImplementedError(
        f'Cannot convert to string format: {hash_format}.')


def string_to_vector(hash_string: str,
                     dtype: str,
                     hash_length: int,
                     hash_format: str,
                     verify_length: bool = True):
    """Convert hash back to vector.

    Args:
        hash_string: The input base64 hash string
        dtype: The data type of the hash
        hash_length: The length of the hash vector
        verify_length: Whether to verify the string length
    """
    assert not verify_length or len(hash_string) == get_string_length(
        hash_length=hash_length, hash_format=hash_format,
        dtype=dtype), 'Incorrect string length for this hash format.'
    if hash_format == 'base64':
        vector_bytes = np.frombuffer(
            base64.b64decode(hash_string),
            dtype='uint8' if dtype in ['bool', 'uint8'] else dtype)
    elif hash_format == 'hex':
        vector_bytes = np.frombuffer(
            bytearray.fromhex(hash_string),
            dtype='uint8' if dtype in ['bool', 'uint8'] else dtype)
    else:
        raise NotImplementedError(
            f'Cannot convert to string format: {hash_format}')
    if dtype == 'uint8':
        return vector_bytes[:hash_length]
    if dtype == 'float32':
        return vector_bytes[:hash_length]
    if dtype == 'bool':
        return np.unpackbits(vector_bytes)[:hash_length].astype('bool')
    raise NotImplementedError(f'Cannot convert hash of type {dtype}.')


def to_image_array(image: ImageInputType, require_color=True):
    if isinstance(image, np.ndarray):
        assert image.flags['C_CONTIGUOUS'], (
            'Provided arrays must be contiguous to avoid '
            'erroneous results when arrays are passed to '
            'underlying libraries. This can be achieved using'
            'np.ascontiguousarray(image)')
        assert not require_color or (len(image.shape) == 3
                                     and image.shape[-1] == 3), (
                                         'Provided images must be RGB images.')
        return image
    return read(image)


def get_common_framerates(id_rates: dict):
    """Compute an optimal set of framerates for a list
    of framerates. Optimal here means that reading the video
    at each of the framerates will allow one to collect all
    of the frames required with the smallest possible number of
    frames decoded.

    For example, consider if we need to read a video at
    3 fps, 5 fps, 1 fps and 0.5 fps. We could read the video
    4 times (once per framerate). But a more optimal approach
    is to read the video only twice, once at 3 frames per second
    and another time at 5 frames per second. For the 1 fps hasher,
    we simply pass every 3rd frame of the 3 fps pass. For the
    0.5 fps hasher, we pass every 6th frame of the 3 fps pass. So
    if you pass this function {A: 3, B: 5, C: 1, D: 0.5}, you will
    get back {3: [A, C, D], 5: C}.

    Args:
        id_rates: A dictionary with IDs as keys and frame rates as values.

    Returns:
        rate_ids: A dictionary with framerates as keys and a list of
            ids as values.
    """

    def partition(collection):
        """This function taken from
        https://stackoverflow.com/questions/19368375/set-partitions-in-python/30134039#30134039
        """
        if len(collection) == 1:
            yield [collection]
            return

        first = collection[0]
        for smaller in partition(collection[1:]):
            # insert `first` in each of the subpartition's subsets
            for n, subset in enumerate(smaller):
                yield smaller[:n] + [[first] + subset] + smaller[n + 1:]
            # put `first` in its own subset
            yield [[first]] + smaller

    framerates = list(id_rates.values())
    factor = 2 * 3 * 5 * 7 * 11 * 60 * 60
    assert min(framerates
               ) >= 1 / factor, 'Framerates must be at least 1 frame per hour.'
    best_frame_count = np.inf
    best_grouping: typing.Optional[typing.List] = None
    best_frame_rates: typing.Optional[typing.List] = None

    # We try every possible grouping of framerates to minimize the number
    # of frames we decode. There is likely a better way to do this,
    # but this seems to do the job for now.
    for grouping in partition(list(set(framerates))):
        current_frame_rates = [
            # pylint: disable=no-member
            functools.reduce(np.lcm,
                             (np.array(group) * factor).round().astype(int)) /
            factor for group in grouping
        ]
        current_frame_count = sum(current_frame_rates)
        if current_frame_count < best_frame_count:
            best_frame_count = current_frame_count
            best_frame_rates = current_frame_rates
            best_grouping = grouping

    assert best_frame_rates is not None
    assert best_grouping is not None
    return {
        framerate:
        tuple(name for name, rate in id_rates.items() if rate in group)
        for framerate, group in zip(best_frame_rates, best_grouping)
    }


def get_isometric_transforms(image: ImageInputType, require_color=True):
    image = to_image_array(image, require_color=require_color)
    return dict(
        r0=image,
        fv=np.ascontiguousarray(image[::-1, :]),
        fh=np.ascontiguousarray(image[:, ::-1]),
        r180=np.ascontiguousarray(image[::-1, ::-1]),
        r90=np.ascontiguousarray(image.transpose(1, 0, 2)[::-1, :, :]),
        r90fv=np.ascontiguousarray(image.transpose(1, 0, 2)),
        r90fh=np.ascontiguousarray(image.transpose(1, 0, 2)[::-1, ::-1]),
        r270=np.ascontiguousarray(image.transpose(1, 0, 2)[:, ::-1]))


def get_isometric_dct_transforms(dct: np.ndarray):
    # pylint: disable=invalid-name
    T1 = np.empty_like(dct)
    T1[::2] = 1
    T1[1::2] = -1

    # pylint: disable=invalid-name
    T2 = np.empty_like(dct)
    T2[::2, ::2] = 1
    T2[1::2, 1::2] = 1
    T2[::2, 1::2] = -1
    T2[1::2, ::2] = -1
    return dict(
        r0=dct,
        fv=dct * T1,
        fh=dct * T1.T,
        r180=dct * T2,
        r90=dct.T * T1,
        r90fv=dct.T,
        r90fh=dct.T * T2,
        r270=dct.T * T1.T)


def read(filepath_or_buffer: ImageInputType, timeout=None):
    """Read a file into an image object

    Args:
        filepath_or_buffer: The path to the file or any object
            with a `read` method (such as `io.BytesIO`)
        timeout: If filepath_or_buffer is a URL, the timeout to
            use for making the HTTP request.
    """
    if PIL is not None and isinstance(filepath_or_buffer, PIL.Image.Image):
        return np.array(filepath_or_buffer.convert("RGB"))
    if isinstance(filepath_or_buffer, (io.BytesIO, client.HTTPResponse)):
        image = np.asarray(
            bytearray(filepath_or_buffer.read()), dtype=np.uint8)
        image = cv2.imdecode(image, cv2.IMREAD_UNCHANGED)
    elif (isinstance(filepath_or_buffer, str)
          and validators.url(filepath_or_buffer)):
        return read(request.urlopen(filepath_or_buffer, timeout=timeout))
    else:
        if not os.path.isfile(filepath_or_buffer):
            raise FileNotFoundError('Could not find image at path: ' +
                                    filepath_or_buffer)
        image = cv2.imread(filepath_or_buffer)
    if image is None:
        raise ValueError(f'An error occurred reading {filepath_or_buffer}.')
    # We use cvtColor here instead of just ret[..., ::-1]
    # in order to ensure that we provide a contiguous
    # array for later processing. Some hashers use ctypes
    # to pass the array and non-contiguous arrays can lead
    # to erroneous results.
    return cv2.cvtColor(image, cv2.COLOR_BGR2RGB)


def _get_frame_types(filepath):
    """Get the frame types for the frames in a video.

    Args:
        filepath: Path to the target file

    Returns:
        A list of dictionaries with pict_type
        (values are 'I', 'P', or 'B') and
        coded_picture_number (which represents the
        frame).
    """
    args = [
        'ffprobe', '-select_streams', 'v', '-i', filepath, '-print_format',
        'json', '-show_entries', 'frame=pict_type,coded_picture_number'
    ]
    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()
    if p.returncode != 0:
        raise ValueError("{out}: {err}".format(out=str(out), err=str(err)))
    frames = json.loads(out.decode('utf-8'))['frames']
    frames.sort(key=lambda f: f['coded_picture_number'])
    return frames


def _get_keyframes(filepath):
    """Get the keyframes for a video.

    Args:
        filepath: Path to the target file

    Returns:
        A list of frame indexes.
    """
    args = [
        'ffprobe', '-select_streams', 'v', '-i', filepath, '-print_format',
        'json', '-show_entries', 'frame=pict_type,coded_picture_number'
    ]
    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()
    if p.returncode != 0:
        raise ValueError("{out}: {err}".format(out=str(out), err=str(err)))
    data = json.loads(out.decode('utf-8'))['frames']
    frames = [f['coded_picture_number'] for f in data if f['pict_type'] == 'I']
    frames = list(set(frames))
    frames.sort()
    return frames


# pylint: disable=too-many-branches,too-many-locals,too-many-statements
def read_video_to_generator(
        filepath,
        frames_per_second: typing.Optional[typing.Union[str, float]] = None,
        errors='raise',
        max_duration: float = None,
        max_size: int = None):
    # pylint: disable=no-member
    if cv2.__version__ < '4.1.1' and filepath.lower().endswith('gif'):
        message = 'Versions of OpenCV < 4.1.1 may read GIF files improperly. Upgrade recommended.'
        if errors == 'raise':
            raise ValueError(message)
        warnings.warn(message=message)

    if not os.path.isfile(filepath):
        raise FileNotFoundError(f'Could not find {filepath}.')
    cap = cv2.VideoCapture(filename=filepath, apiPreference=cv2.CAP_FFMPEG)
    try:
        # The purpose of the following block is largely to create a
        # frame_indexes (iterator or list) that indicates which
        # frames we should be returning to the user and then
        # yielding those frames as we come across them.
        file_frames_per_second = cap.get(cv2.CAP_PROP_FPS)
        if file_frames_per_second == 0:
            if errors == "raise":
                raise ValueError("Video file has framerate of 0fps.")
            # The known case where this occurs is for GIFs, where
            # 0 fps is typically inferred as 10 fps.
            file_frames_per_second = 10
            if errors == "warn":
                warnings.warn(
                    message=
                    "Video file has framerate of 0 fps. Guessing framerate of 10fps."
                )
        if frames_per_second is None:
            frames_per_second = file_frames_per_second
        seconds_between_desired_frames = None if (
            frames_per_second is not None
            and isinstance(frames_per_second,
                           str)) else 1 / frames_per_second  # type: ignore
        seconds_between_grabbed_frames = 1 / file_frames_per_second
        grabbed_frame_count = 0
        if frames_per_second == 'keyframes':
            frame_indexes: typing.Union[range, typing.List[int], typing.
                                        Iterator[int]] = _get_keyframes(
                                            filepath)
            # The repeat flag is used to handle the case where the
            # desired sampling rate is higher than the file's frame
            # rate. In this case, we will need to repeat frames in
            # order to provide the least-surprising behavior that
            # we can.
            repeat = False
        else:
            frame_indexes = itertools.count(
                0, max(1, file_frames_per_second / frames_per_second))
            repeat = file_frames_per_second < frames_per_second
        input_width = cap.get(cv2.CAP_PROP_FRAME_WIDTH)
        input_height = cap.get(cv2.CAP_PROP_FRAME_HEIGHT)
        if max_size is not None:
            scale = min(max_size / max(input_width, input_height), 1)
        else:
            scale = 1
        target_size: typing.Optional[typing.Tuple[int, int]]
        if scale < 1:
            target_size = (int(scale * input_width), int(scale * input_height))
        else:
            target_size = None
        for frame_index in frame_indexes:
            while grabbed_frame_count < frame_index:
                # We need to skip this frame.
                success = cap.grab()
                if not success:
                    break
                grabbed_frame_count += 1
            success, frame = cap.read()
            grabbed_frame_count += 1
            if not success:
                # The video is over or an error has occurred.
                break
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            if target_size is not None:
                frame = cv2.resize(
                    frame, target_size, interpolation=cv2.INTER_NEAREST)
            current_timestamp = frame_index / file_frames_per_second
            yield frame, grabbed_frame_count - 1, current_timestamp
            if max_duration is not None and current_timestamp > max_duration:
                break
            if repeat:
                next_desired_timestamp = current_timestamp + seconds_between_desired_frames
                next_timestamp = current_timestamp + seconds_between_grabbed_frames
                while next_desired_timestamp < next_timestamp:
                    yield (frame, grabbed_frame_count - 1,
                           next_desired_timestamp)
                    next_desired_timestamp += seconds_between_desired_frames
    # pylint: disable=broad-except
    except Exception as e:
        if errors not in ['warn', 'ignore']:
            raise e
        if errors == 'warn':
            warnings.warn(
                message=
                f'An error occurred while reading {filepath}. Processing may be truncated.'
            )
    finally:
        cap.release()


def read_video_into_queue(*args, video_queue, terminate, **kwargs):
    # We're inside a thread now and the queue is being read elsewhere.
    try:
        for frame, frame_index, timestamp in read_video_to_generator(
                *args, **kwargs):
            if not terminate.isSet():
                video_queue.put((frame, frame_index, timestamp))
            else:
                break
    finally:
        video_queue.put((None, None, None))


# pylint: disable=too-many-arguments
def read_video(
        filepath,
        frames_per_second: typing.Optional[typing.Union[str, float]] = None,
        max_queue_size=128,
        use_queue=True,
        errors='raise',
        **kwargs):
    """Provides a generator of RGB frames, frame indexes, and timestamps from a
    video. This function requires you to have installed ffmpeg. All other
    arguments passed to read_video_to_generator.

    Args:
        filepath: Path to the video file
        frames_per_second: How many frames to provide for
            each second of video. If None, all frames
            are provided. If frames_per_second is "keyframes",
            we use ffmpeg to select I frames from the video.
        max_queue_size: The maximum number of frames to load in the queue
        use_queue: Whether to use a queue of frames during processing
        max_duration: The maximum length of the video to hash.
        max_size: The maximum size of frames to queue
        errors: Whether to 'raise', 'warn', or 'ignore' errors

    Yields:
        (frame, frame_index, timestamp) tuples
    """
    if use_queue:
        video_queue = queue.Queue(
            maxsize=max_queue_size
        )  # type: queue.Queue[typing.Tuple[np.ndarray, int, float]]
        terminate = threading.Event()
        thread = threading.Thread(
            target=read_video_into_queue,
            kwargs={
                'frames_per_second': frames_per_second,
                'video_queue': video_queue,
                'filepath': filepath,
                'errors': errors,
                'terminate': terminate,
                **kwargs
            })
        thread.start()
        try:
            while True:
                frame, frame_index, timestamp = video_queue.get()
                video_queue.task_done()
                if frame is None:
                    break
                yield (frame, frame_index, timestamp)
        finally:
            # Set the termination flag for the
            # background thread.
            terminate.set()
            try:
                # Unblock the thread, in the event
                # that it is waiting.
                video_queue.get_nowait()

                # Do it twice for the edge case
                # where the queue is completely
                # full and the end sentinel is
                # blocking.
                video_queue.get_nowait()
            except queue.Empty:
                # It doesn't matter if it's empty.
                pass
            # Wait for the background thread to terminate.
            thread.join()
    else:
        for frame, frame_index, timestamp in read_video_to_generator(
                filepath=filepath,
                frames_per_second=frames_per_second,
                errors=errors,
                **kwargs):
            yield (frame, frame_index, timestamp)


def compute_synchronized_video_hashes(filepath: str,
                                      hashers: dict,
                                      framerates=None,
                                      hash_format='base64',
                                      use_queue=True):
    """Compute the video hashes for a group of hashers with synchronized
    frame processing wherever possible.

    Args:
        filepath: Path to video file.
        hashers: A dictionary mapping hasher names to video hasher objects
        hash_format: The format in which to return the hashes
        use_queue: Whether to use queued video frames
    """
    if framerates is None:
        framerates = get_common_framerates({
            k: h.frames_per_second
            for k, h in hashers.items() if h.frames_per_second is not None
        })
    else:
        assert all(
            any(hasher_name in hasher_names
                for hasher_names in framerates.values())
            for hasher_name, hasher in hashers.items()
            if hasher.frames_per_second is not None
        ), 'Provided framerates do not have an entry for all required hashers.'

    results = {
        hasher_name: {
            'state':
            None,
            'hash':
            None,
            'relative_framerate':
            next(framerate / hasher.frames_per_second
                 for framerate, hasher_names in framerates.items()
                 if hasher_name in hasher_names)
        }
        for hasher_name, hasher in hashers.items()
        if hasher.frames_per_second is not None
    }
    for current_framerate, current_hasher_names in framerates.items():
        for frame_index, (frame, grabbed_frame_index,
                          frame_timestamp) in enumerate(
                              read_video(
                                  filepath=filepath,
                                  frames_per_second=current_framerate,
                                  use_queue=use_queue)):
            for hasher_name in current_hasher_names:
                config = results[hasher_name]
                hasher = hashers[hasher_name]
                assert config['relative_framerate'] is not None
                if frame_index % config['relative_framerate'] == 0:
                    config['state'] = hasher.process_frame(
                        frame=frame,
                        frame_index=grabbed_frame_index,
                        frame_timestamp=frame_timestamp,
                        state=config['state'])
        for hasher_name in current_hasher_names:
            config = results[hasher_name]
            hasher = hashers[hasher_name]
            current_hash = hasher.hash_from_final_state(state=config['state'])
            if hash_format == 'vector':
                config['hash'] = current_hash
            else:
                if not hasher.returns_multiple:
                    config['hash'] = hasher.vector_to_string(
                        current_hash, hash_format=hash_format)
                else:
                    config['hash'] = [
                        hasher.vector_to_string(h, hash_format=hash_format)
                        for h in current_hash
                    ]
            config['state'] = None
    hashes = {
        hasher_name: config['hash']
        for hasher_name, config in results.items()
    }
    for hasher_name, hasher in hashers.items():
        if hasher.frames_per_second is None:
            # This is a custom hasher that we just pass a video path to.
            hashes[hasher_name] = hasher.compute(filepath)
    return hashes


def unletterbox(image) -> typing.Optional[
        typing.Tuple[typing.Tuple[int, int], typing.Tuple[int, int]]]:
    """Obtain bounds on an image that remove the black bars
    on the top, right, bottom, and left side of an image.

    Args:
        image: The image from which to remove letterboxing.

    Returns:
        A pair of coordinates bounds of the form (x1, x2)
        and (y1, y2) representing the left, right, top, and
        bottom bounds.
    """
    adj = image.mean(axis=2) > 2
    if adj.all():
        bounds = (0, image.shape[1] + 1), (0, image.shape[0])
    else:
        y = np.where(adj.sum(axis=1) > 0.1 * image.shape[0])[0]
        x = np.where(adj.sum(axis=0) > 0.1 * image.shape[1])[0]
        if len(y) <= 1 or len(x) <= 1:
            return None
        x1, x2 = x[[0, -1]]
        y1, y2 = y[[0, -1]]
        bounds = (x1, x2 + 1), (y1, y2 + 1)
    return bounds



import concurrent.futures
import typing
import uuid
import os

import tqdm
import pandas as pd

from ..tools import flatten
from ..hashers import VideoHasher, tools
from .common import BenchmarkDataset, BenchmarkTransforms, BenchmarkHashes


def _process_row(row, hashers, framerates):
    error = None
    try:
        assert not pd.isnull(row['filepath']), 'No filepath provided.'
        hashes = tools.compute_synchronized_video_hashes(
            filepath=row['filepath'],
            hashers=hashers,
            framerates=framerates,
            hash_format='base64')
    # pylint: disable=broad-except
    except Exception as exception:
        error = str(exception)
        hashes = {
            hasher_name: [None] if hasher.returns_multiple else None
            for hasher_name, hasher in hashers.items()
        }
    base_dict = {
        'guid': row['guid'],
        'filepath': row['filepath'],
        'error': error,
        'category': row['category'],
        'transform_name': row['transform_name'],
        'input_filepath': row['input_filepath']
    }
    hash_dicts = []
    for hasher_name, hasher in hashers.items():
        base_hash_dict = {
            'hasher_name': hasher_name,
            'hasher_dtype': hasher.dtype,
            'hasher_distance_metric': hasher.distance_metric,
            'hasher_hash_length': hasher.hash_length,
        }
        if not hasher.returns_multiple:
            hash_dicts.append({
                **{
                    'hash': hashes[hasher_name],
                },
                **base_hash_dict
            })
        else:
            for hash_value in hashes[hasher_name]:
                hash_dicts.append({
                    **{
                        'hash': hash_value,
                    },
                    **base_hash_dict
                })
    return [{**hash_dict, **base_dict} for hash_dict in hash_dicts]


class BenchmarkVideoDataset(BenchmarkDataset):
    def transform(self,
                  transforms: typing.Dict[str, typing.Callable],
                  storage_dir: str,
                  errors: str = "raise"):
        """Prepare files to be used as part of benchmarking run.

        Args:
            transforms: A dictionary of transformations. The only required
                key is `noop` which determines how the original, untransformed
                video is saved. Each transform should be a callable function with
                that accepts an `input_filepath` and `output_filepath` argument and
                it should return the `output_filepath` (which may have a different
                extension appended by the transform function).
            storage_dir: A directory to store all the videos along with
                their transformed counterparts.
            errors: How to handle errors reading files. If "raise", exceptions are
                raised. If "warn", the error is printed as a warning.

        Returns:
            transforms: A BenchmarkVideoTransforms object
        """
        assert 'noop' in transforms, 'You must provide a no-op transform.'

        os.makedirs(storage_dir, exist_ok=True)

        files = self._df.copy()
        files['guid'] = [uuid.uuid4() for n in range(len(files))]

        def apply_transform_to_file(input_filepath, guid, transform_name,
                                    category):
            if input_filepath is None:
                # This can happen if the noop transform did not yield
                # a file. We don't want to drop the records so we
                # keep them.
                return {
                    'guid': guid,
                    'error': 'No source file provided',
                    'transform_name': transform_name,
                    'input_filepath': input_filepath,
                    'filepath': None,
                    'category': category
                }
            try:
                output_filepath = transforms[transform_name](
                    input_filepath,
                    output_filepath=os.path.join(storage_dir,
                                                 f'{guid}_{transform_name}'))
                error = None
            # pylint: disable=invalid-name,broad-except
            except Exception as e:
                output_filepath = None
                error = str(e)
            return {
                'guid': guid,
                'error': error,
                'transform_name': transform_name,
                'input_filepath': input_filepath,
                'filepath': output_filepath,
                'category': category
            }

        def apply_transform_to_files(files, transform_name):
            return pd.DataFrame.from_records([
                apply_transform_to_file(
                    input_filepath=row['filepath'],
                    guid=row['guid'],
                    transform_name=transform_name,
                    category=row['category']) for _, row in tqdm.tqdm(
                        files.iterrows(),
                        desc=f'Creating files for {transform_name}',
                        total=len(files))
            ])

        results = [apply_transform_to_files(files, transform_name='noop')]
        for transform_name in transforms.keys():
            if transform_name == 'noop':
                continue
            results.append(
                apply_transform_to_files(
                    results[0], transform_name=transform_name))
        benchmark_transforms = BenchmarkVideoTransforms(
            df=pd.concat(results, axis=0, ignore_index=True))
        benchmark_transforms.save(storage_dir)
        return benchmark_transforms


class BenchmarkVideoTransforms(BenchmarkTransforms):
    expected_columns = [
        'filepath', 'category', 'transform_name', 'input_filepath', 'guid',
        'error'
    ]

    def compute_hashes(self,
                       hashers: typing.Dict[str, VideoHasher],
                       max_workers: int = 5) -> BenchmarkHashes:
        """Compute hashes for a series of files given some set of hashers.

        Args:
            hashers: A dictionary of hashers.
            max_workers: Maximum number of workers for parallel hash
                computation.

        Returns:
            hashes: A BenchmarkHashes object.
        """
        id_rates = {
            hasher_name: hasher.frames_per_second
            for hasher_name, hasher in hashers.items()
            if hasher.frames_per_second is not None
        }
        if id_rates:
            framerates = tools.get_common_framerates({
                hasher_name: hasher.frames_per_second
                for hasher_name, hasher in hashers.items()
                if hasher.frames_per_second is not None
            })
        else:
            framerates = {}

        with concurrent.futures.ProcessPoolExecutor(
                max_workers=max_workers) as executor:
            futures = [
                executor.submit(
                    _process_row,
                    row=row,
                    framerates=framerates,
                    hashers=hashers) for index, row in self._df.iterrows()
            ]
            return BenchmarkHashes(
                pd.DataFrame.from_records(
                    flatten([
                        future.result() for future in tqdm.tqdm(
                            concurrent.futures.as_completed(futures),
                            desc='Computing hashes.',
                            total=len(self._df))
                    ])))



# pylint: disable=line-too-long

import cv2
import numpy as np

from ..hasher import ImageHasher


class OpenCVHasher(ImageHasher):  # pylint: disable=abstract-method
    allow_parallel = False

    def __init__(self):
        if not hasattr(cv2, 'img_hash'):
            raise Exception(
                'You do not appear to have opencv-contrib installed. It is required for pure OpenCV hashers.'  # pylint: disable=line-too-long
            )


class MarrHildreth(OpenCVHasher):
    """A wrapper around OpenCV's Marr-Hildreth hash.
    See `paper <https://www.phash.org/docs/pubs/thesis_zauner.pdf>`_ for details."""

    dtype = 'bool'
    distance_metric = 'hamming'
    hash_length = 576

    def __init__(self):
        super().__init__()
        self.hasher = cv2.img_hash_MarrHildrethHash.create()  # pylint: disable=no-member

    def _compute(self, image):
        return np.unpackbits(self.hasher.compute(image)[0])


class ColorMoment(OpenCVHasher):
    """A wrapper around OpenCV's Color Moments hash.
    See `paper <https://www.phash.org/docs/pubs/thesis_zauner.pdf>`_ for details."""

    dtype = 'float32'
    distance_metric = 'euclidean'
    hash_length = 42

    def __init__(self):
        super().__init__()
        self.hasher = cv2.img_hash_ColorMomentHash.create()  # pylint: disable=no-member

    def _compute(self, image):
        return 10000 * self.hasher.compute(image)[0]


class BlockMean(OpenCVHasher):
    """A wrapper around OpenCV's Block Mean hash.
    See `paper <https://www.phash.org/docs/pubs/thesis_zauner.pdf>`_ for details."""
    dtype = 'bool'
    distance_metric = 'hamming'
    hash_length = 968

    def __init__(self):
        super().__init__()
        self.hasher = cv2.img_hash_BlockMeanHash.create(1)  # pylint: disable=no-member

    def _compute(self, image):
        # https://stackoverflow.com/questions/54762896/why-cv2-norm-hamming-gives-different-value-than-actual-hamming-distance
        return np.unpackbits(self.hasher.compute(image)[0])



from ..hasher import VideoHasher, ImageHasher
from .. import tools


class FramewiseHasher(VideoHasher):
    """A hasher that simply returns frame-wise hashes at some
    regular interval with some minimum inter-frame distance threshold."""

    returns_multiple = True

    def __init__(self,
                 frame_hasher: ImageHasher,
                 interframe_threshold: float,
                 frames_per_second: int = 15,
                 quality_threshold: float = None):
        self.hash_length = frame_hasher.hash_length
        self.frames_per_second = frames_per_second
        self.frame_hasher = frame_hasher
        self.distance_metric = frame_hasher.distance_metric
        self.dtype = frame_hasher.dtype
        self.interframe_threshold = interframe_threshold
        self.quality_threshold = quality_threshold

    # pylint: disable=unused-argument
    def process_frame(self, frame, frame_index, frame_timestamp, state=None):
        if self.quality_threshold is not None:
            current, quality = self.frame_hasher.compute_with_quality(
                frame, hash_format='vector')
            if quality < self.quality_threshold:
                return state or {'previous': None, 'hashes': []}
        else:
            current = self.frame_hasher.compute(frame, hash_format='vector')
        if state is None or state['previous'] is None:
            # We keep a separate reference to the previous hash instead of using
            # the last entry in the hashes list because `compute_batches` may
            # clear the hashes list but we still want to be able to compare
            # the final entry.
            state = {
                'previous': current,
                'hashes': [current],
            }
        else:
            if self.frame_hasher.compute_distance(
                    current, state['previous']) > self.interframe_threshold:
                state['hashes'].append(current)
        return state

    def compute_batches(self,
                        filepath: str,
                        batch_size: int,
                        errors='raise',
                        hash_format='base64'):
        """Compute hashes for a video in batches.

        Args:
            filepath: Path to video file
            batch_size: The batch size to use for returning hashes
            errors: One of "raise", "ignore", or "warn". Passed
                to perception.hashers.tools.read_video.
            hash_format: The format in which to return hashes
        """

        def format_batch(hashes):
            return [
                self.vector_to_string(vector, hash_format=hash_format)
                if hash_format != 'vector' else vector for vector in hashes
            ]

        state = None
        for frame, frame_index, frame_timestamp in tools.read_video(
                filepath=filepath,
                frames_per_second=self.frames_per_second,
                errors=errors):
            state = self.process_frame(
                frame=frame,
                frame_index=frame_index,
                frame_timestamp=frame_timestamp,
                state=state)
            if state is not None and len(state['hashes']) > batch_size:
                yield format_batch(state['hashes'])
                state['hashes'] = []
        if state is not None and state['hashes']:
            yield format_batch(state['hashes'])

    def hash_from_final_state(self, state):
        if state is None:
            return []
        return state['hashes']



from .image import BenchmarkImageDataset, BenchmarkImageTransforms
from .video import BenchmarkVideoDataset, BenchmarkVideoTransforms
from .common import BenchmarkHashes
from . import video_transforms
from . import video
from . import image



import scipy.fftpack
import numpy as np
import cv2

from ..hasher import ImageHasher
from .. import tools


class PHash(ImageHasher):
    """Also known as the DCT hash, a hash based on discrete cosine transforms of images.
    See `complete paper <https://www.phash.org/docs/pubs/thesis_zauner.pdf>`_ for
    details. Implementation based on that of
    `ImageHash <https://github.com/JohannesBuchner/imagehash>`_.

    Args:
        hash_size: The number of DCT elements to retain (the hash length
            will be hash_size * hash_size).
        highfreq_factor: The multiple of the hash size to resize the input
            image to before computing the DCT.
        exclude_first_term: WHether to exclude the first term of the DCT
        freq_shift: The number of DCT low frequency elements to skip.
    """
    distance_metric = 'hamming'
    dtype = 'bool'

    def __init__(self,
                 hash_size=8,
                 highfreq_factor=4,
                 exclude_first_term=False,
                 freq_shift=0):
        assert hash_size >= 2, 'Hash size must be greater than or equal to 2'
        assert freq_shift <= highfreq_factor * hash_size - hash_size, \
            'Frequency shift is too large for this hash size / highfreq_factor combination.'
        self.hash_size = hash_size
        self.highfreq_factor = highfreq_factor
        self.exclude_first_term = exclude_first_term
        self.hash_length = hash_size * hash_size
        self.freq_shift = freq_shift
        if exclude_first_term:
            self.hash_length -= 1

    def _compute_dct(self, image):
        img_size = self.hash_size * self.highfreq_factor
        image = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
        image = cv2.resize(
            image, dsize=(img_size, img_size), interpolation=cv2.INTER_AREA)
        dct = scipy.fftpack.dct(scipy.fftpack.dct(image, axis=0), axis=1)
        return dct[self.freq_shift:self.hash_size + self.freq_shift, self.
                   freq_shift:self.hash_size + self.freq_shift]

    # pylint: disable=no-self-use
    def _dct_to_hash(self, dct):
        dct = dct.flatten()
        if self.exclude_first_term:
            dct = dct[1:]
        return dct > np.median(dct)

    def _compute(self, image):
        dct = self._compute_dct(image)
        return self._dct_to_hash(dct)

    def _compute_isometric(self, image):
        return {
            transform_name: self._dct_to_hash(dct)
            for transform_name, dct in tools.get_isometric_dct_transforms(
                self._compute_dct(image)).items()
        }


class PHashF(PHash):
    """A real-valued version of PHash. It
    returns the raw 32-bit floats in the DCT.
    For a more compact approach, see PHashU8."""
    dtype = 'float32'
    distance_metric = 'euclidean'

    def _dct_to_hash(self, dct):
        dct = dct.flatten()
        if self.exclude_first_term:
            dct = dct[1:]
        if (dct == 0).all():
            return None
        return dct


class PHashU8(PHash):
    """A real-valued version of PHash. It
    uses minimum / maximum scaling to convert
    DCT values to unsigned 8-bit integers (more
    compact than the 32-bit floats used by PHashF at
    the cost of precision)."""
    dtype = 'uint8'
    distance_metric = 'euclidean'

    def _dct_to_hash(self, dct):
        dct = dct.flatten()
        if self.exclude_first_term:
            dct = dct[1:]
        if (dct == 0).all():
            return None
        min_value = dct.min()
        max_value = dct.max()
        dct = np.uint8(255 * (dct - min_value) / (max_value - min_value))
        return dct



# pylint: disable=invalid-name
from abc import ABC
import logging
import tempfile
import typing
import itertools
import warnings
import zipfile
import shutil
import uuid
import os

import matplotlib.pyplot as plt
from scipy import spatial, stats
import pandas as pd
import numpy as np
import tqdm

from ..hashers.tools import compute_md5, string_to_vector
try:
    from . import extensions  # type: ignore
except ImportError:
    warnings.warn(
        'C extensions were not built. Some metrics will be computed more slowly. '
        'Please install from wheels or set up a compiler prior to installation '
        'from source to use extensions.')
    extensions = None

log = logging.getLogger(__name__)


def create_mask(transformed_guids, noop_guids):
    """Given a list of transformed guids and noop guids,
    computes an MxN array indicating whether noop n has the same guid
    as transform m. Used for applying a mask to a distance matrix
    for efficient computation of recall at different thresholds.

    Args:
        transformed_guids: An iterable of transformed guids
        noop: An iterable of noop guids

    Returns:
        An boolean array of shape
        `(len(transformed_guids), len(transformed_noops))`
    """
    n_noops = len(noop_guids)
    previous_guid = None
    start = None
    end = 0
    mask = np.zeros((len(transformed_guids), len(noop_guids)), dtype='bool')
    for current_guid, row in zip(transformed_guids, mask):
        if previous_guid is None or current_guid != previous_guid:
            start = end
            end = start + next(
                (other_index
                 for other_index, guid in enumerate(noop_guids[start:])
                 if guid != current_guid), n_noops)
            previous_guid = current_guid
        row[start:end] = True
    return mask


def compute_threshold_fpr_recall(pos, neg, fpr_threshold=0.001):
    # Sort both arrays according to the positive distance
    neg = neg[pos.argsort()]
    pos = pos[pos.argsort()]

    # Compute false-positive rate for every value in pos
    tp = np.arange(1, len(pos) + 1)
    fp = np.array([(neg <= t).sum() for t in pos])
    fpr = fp / (tp + fp)

    # Choose the optimal threshold
    bad_thresholds = pos[fpr > fpr_threshold]

    # pylint: disable=len-as-condition
    if len(bad_thresholds) > 0:
        optimal_threshold = bad_thresholds[0]
        recovered = (pos < optimal_threshold).sum()
        if recovered == 0:
            optimal_fpr = 0
        else:
            optimal_fpr = fpr[pos < optimal_threshold].max()
        optimal_recall = round(100 * recovered / len(pos), 3)
    else:
        optimal_fpr = 0
        optimal_threshold = pos.max()
        optimal_recall = 100
    return optimal_threshold, optimal_fpr, optimal_recall


class Filterable(ABC):
    _df: pd.DataFrame
    expected_columns: typing.List

    def __init__(self, df):
        # pylint: disable=no-member
        assert sorted(df.columns) == sorted(
            self.expected_columns
        ), f'Column mismatch: Expected {sorted(self.expected_columns)}, found {sorted(df.columns)}.'
        # pylint: enable=no-member
        self._df = df

    @property
    def categories(self):
        """The categories included in the dataset"""
        return self._df['category'].unique()

    def filter(self, **kwargs):
        """Obtain a new dataset filtered with the given
        keyword arguments."""
        df = self._df.copy()
        for field, included in kwargs.items():
            existing = self._df[field].unique()
            if not all(inc in existing for inc in included):
                message = 'Did not find {missing} in column {field} dataset.'.format(
                    missing=', '.join(
                        [str(inc) for inc in included if inc not in existing]),
                    field=field)
                warnings.warn(message, UserWarning)
            df = df[df[field].isin(included)]
        return self.__class__(df.copy())


class Saveable(Filterable):
    @classmethod
    def load(cls,
             path_to_zip_or_directory: str,
             storage_dir: str = None,
             verify_md5=True):
        """Load a dataset from a ZIP file or directory.

        Args:
            path_to_zip_or_directory: Pretty self-explanatory
            storage_dir: If providing a ZIP file, where to extract
                the contents. If None, contents will be extracted to
                a folder with the same name as the ZIP file in the
                same directory as the ZIP file.
            verify_md5: Verify md5s when loading
        """

        # Load index whether from inside ZIP file or from directory.
        if os.path.splitext(path_to_zip_or_directory)[1] == '.zip':
            if storage_dir is None:
                storage_dir = os.path.join(
                    os.path.dirname(os.path.abspath(path_to_zip_or_directory)),
                    os.path.splitext(
                        os.path.basename(path_to_zip_or_directory))[0])
                os.makedirs(storage_dir, exist_ok=True)
            with zipfile.ZipFile(path_to_zip_or_directory, 'r') as z:
                # Try extracting only the index at first so we can
                # compare md5.
                z.extract('index.csv', os.path.join(storage_dir))
                index = pd.read_csv(os.path.join(storage_dir, 'index.csv'))
                index['filepath'] = index['filename'].apply(
                    lambda fn: os.path.join(storage_dir, fn) if not pd.isnull(fn) else None
                )
                if index['filepath'].apply(os.path.isfile).all() and (
                        not verify_md5
                        or all(row['md5'] == compute_md5(row['filepath']))
                        for _, row in tqdm.tqdm(
                            index.iterrows(), desc='Checking cache.')):
                    log.info(
                        'Found all files already extracted. Skipping extraction.'
                    )
                    verify_md5 = False
                else:
                    z.extractall(storage_dir)
        else:
            assert storage_dir is None, 'Storage directory only valid if path is to ZIP file.'
            index = pd.read_csv(
                os.path.join(path_to_zip_or_directory, 'index.csv'))
            index['filepath'] = index['filename'].apply(
                lambda fn: os.path.join(path_to_zip_or_directory, fn) if not pd.isnull(fn) else None
            )

        if verify_md5:
            assert all(
                row['md5'] == compute_md5(row['filepath'])
                for _, row in tqdm.tqdm(
                    index.iterrows(),
                    desc='Performing final md5 integrity check.',
                    total=len(index.index))), 'An md5 mismatch has occurred.'
        return cls(index.drop(['filename', 'md5'], axis=1))

    def save(self, path_to_zip_or_directory):
        """Save a dataset to a directory or ZIP file.

        Args:
            path_to_zip_or_directory: Pretty self-explanatory
        """
        df = self._df
        assert 'filepath' in df.columns, 'Index dataframe must contain filepath.'

        # Build index using filename instead of filepath.
        index = df.copy()
        index['filename'] = df['filepath'].apply(
            lambda filepath: os.path.basename(filepath) if not pd.isnull(filepath) else None
        )
        if index['filename'].dropna().duplicated().sum() > 0:
            warnings.warn('Changing filenames to UUID due to duplicates.',
                          UserWarning)

            index['filename'] = [
                str(uuid.uuid4()) + os.path.splitext(row['filename'])[1]
                if not pd.isnull(row['filename']) else None
                for _, row in index.iterrows()
            ]
        index['md5'] = [
            compute_md5(filepath) if not pd.isnull(filepath) else None
            for filepath in tqdm.tqdm(
                index['filepath'], desc='Computing md5s.')
        ]

        # Add all files as well as the dataframe index to
        # a ZIP file if path is to ZIP file or to the directory if it is
        # not a ZIP file.
        if os.path.splitext(path_to_zip_or_directory)[1] == '.zip':
            with zipfile.ZipFile(path_to_zip_or_directory, 'w') as f:
                with tempfile.TemporaryFile(mode='w+') as index_file:
                    index.drop(
                        'filepath', axis=1).to_csv(
                            index_file, index=False)
                    index_file.seek(0)
                    f.writestr('index.csv', index_file.read())
                for _, row in tqdm.tqdm(
                        index.iterrows(), desc='Saving files', total=len(df)):
                    if pd.isnull(row['filepath']):
                        #  There was an error associated with this file.
                        continue
                    f.write(row['filepath'], row['filename'])
        else:
            os.makedirs(path_to_zip_or_directory, exist_ok=True)
            index.drop(
                'filepath', axis=1).to_csv(
                    os.path.join(path_to_zip_or_directory, 'index.csv'),
                    index=False)
            for _, row in tqdm.tqdm(
                    index.iterrows(), desc='Saving files', total=len(df)):
                if pd.isnull(row['filepath']):
                    # There was an error associated with this file.
                    continue
                if row['filepath'] == os.path.join(path_to_zip_or_directory,
                                                   row['filename']):
                    # The source file is the same as the target file.
                    continue
                shutil.copy(
                    row['filepath'],
                    os.path.join(path_to_zip_or_directory, row['filename']))


class BenchmarkHashes(Filterable):
    """A dataset of hashes for transformed images. It is essentially
    a wrapper around a `pandas.DataFrame` with the following columns:

    - guid
    - error
    - filepath
    - category
    - transform_name
    - hasher_name
    - hasher_dtype
    - hasher_distance_metric
    - hasher_hash_length
    - hash
    """

    expected_columns = [
        'error', 'filepath', 'hash', 'hasher_name', 'hasher_dtype',
        'hasher_distance_metric', 'category', 'guid', 'input_filepath',
        'transform_name', 'hasher_hash_length'
    ]

    def __init__(self, df: pd.DataFrame):
        super().__init__(df)
        self._metrics: pd.DataFrame = None

    def __add__(self, other):
        return BenchmarkHashes(
            df=pd.concat([self._df, other._df]).drop_duplicates())

    def __radd__(self, other):
        return self.__add__(other)

    @classmethod
    def load(cls, filepath: str):
        return cls(pd.read_csv(filepath))

    def save(self, filepath):
        self._df.to_csv(filepath, index=False)

    # pylint: disable=too-many-locals
    def compute_metrics(self,
                        custom_distance_metrics: dict = None) -> pd.DataFrame:
        if self._metrics is not None:
            return self._metrics
        metrics = []
        hashsets = self._df.sort_values('guid')
        n_dropped = hashsets['hash'].isnull().sum()
        if n_dropped > 0:
            hashsets = hashsets.dropna(subset=['hash'])
            warnings.warn(f'Dropping {n_dropped} invalid / empty hashes.',
                          UserWarning)
        for (hasher_name, transform_name, category), hashset in tqdm.tqdm(
                hashsets.groupby(['hasher_name', 'transform_name',
                                  'category']),
                desc='Computing metrics.'):
            # Note the guid filtering below. We need to include only guids
            # for which we have the transform *and* the guid. One of them
            # may have been dropped due to being invalid.
            noops = hashsets[(hashsets['transform_name'] == 'noop')
                             & (hashsets['hasher_name'] == hasher_name)
                             & (hashsets['guid'].isin(hashset['guid']))]
            hashset = hashset[hashset['guid'].isin(noops['guid'])]
            dtype, distance_metric, hash_length = hashset.iloc[0][[
                'hasher_dtype', 'hasher_distance_metric', 'hasher_hash_length'
            ]]
            n_noops = len(noops.guid)
            n_hashset = len(hashset.guid)
            noop_guids = noops.guid.values
            mask = create_mask(hashset.guid.values, noops.guid.values)
            if distance_metric != 'custom':
                X_trans = np.array(
                    hashset.hash.apply(
                        string_to_vector,
                        hash_length=int(hash_length),
                        dtype=dtype,
                        hash_format='base64').tolist())
                X_noop = np.array(
                    noops.hash.apply(
                        string_to_vector,
                        dtype=dtype,
                        hash_format='base64',
                        hash_length=int(hash_length)).tolist())
                if distance_metric != 'euclidean' or 'int' not in dtype or extensions is None:
                    distance_matrix = spatial.distance.cdist(
                        XA=X_trans, XB=X_noop, metric=distance_metric)
                    distance_to_closest_image = distance_matrix.min(axis=1)
                    distance_to_correct_image = np.ma.masked_array(
                        distance_matrix, np.logical_not(mask)).min(axis=1)
                    distance_matrix_incorrect_image = np.ma.masked_array(
                        distance_matrix, mask)
                    distance_to_incorrect_image = distance_matrix_incorrect_image.min(
                        axis=1)
                    closest_incorrect_guid = noop_guids[
                        distance_matrix_incorrect_image.argmin(axis=1)]
                else:
                    distances, indexes = extensions.compute_euclidean_metrics(
                        X_noop.astype('int32'), X_trans.astype('int32'), mask)
                    distance_to_correct_image = distances[:, 1]
                    distance_to_incorrect_image = distances[:, 0]
                    distance_to_closest_image = distances.min(axis=1)
                    closest_incorrect_guid = [
                        noop_guids[idx] for idx in indexes[:, 0]
                    ]
            else:
                assert (
                    custom_distance_metrics is not None and
                    hasher_name in custom_distance_metrics
                ), \
                    f'You must provide a custom distance metric for {hasher_name}.'
                noops_hash_values = noops.hash.values
                hashset_hash_values = hashset.hash.values
                distance_matrix = np.zeros((n_hashset, n_noops))
                distance_function = custom_distance_metrics[hasher_name]
                for i1, i2 in itertools.product(
                        range(n_hashset), range(n_noops)):
                    distance_matrix[i1, i2] = distance_function(
                        hashset_hash_values[i1], noops_hash_values[i2])
                distance_to_closest_image = distance_matrix.min(axis=1)
                distance_to_correct_image = np.ma.masked_array(
                    distance_matrix, np.logical_not(mask)).min(axis=1)
                distance_matrix_incorrect_image = np.ma.masked_array(
                    distance_matrix, mask)
                distance_to_incorrect_image = distance_matrix_incorrect_image.min(
                    axis=1)
                closest_incorrect_guid = noop_guids[
                    distance_matrix_incorrect_image.argmin(axis=1)]

            metrics.append(
                pd.DataFrame({
                    'guid':
                    hashset['guid'].values,
                    'transform_name':
                    transform_name,
                    'hasher_name':
                    hasher_name,
                    'category':
                    category,
                    'distance_to_closest_correct_image':
                    distance_to_correct_image,
                    'distance_to_closest_incorrect_image':
                    distance_to_incorrect_image,
                    'distance_to_closest_image':
                    distance_to_closest_image,
                    'closest_incorrect_guid':
                    closest_incorrect_guid
                }))
        metrics = pd.concat(metrics)
        self._metrics = metrics
        return metrics

    # pylint: disable=too-many-locals
    def show_histograms(self, grouping=None, fpr_threshold=0.001, **kwargs):
        """Plot histograms for true and false positives, similar
        to https://tech.okcupid.com/evaluating-perceptual-image-hashes-okcupid/
        Additional arguments passed to compute_metrics.

        Args:
            grouping: List of fields to group by. By default, all fields are used
                (category, and transform_name).
        """
        if grouping is None:
            grouping = ['category', 'transform_name']

        metrics = self.compute_metrics(**kwargs)

        hasher_names = metrics['hasher_name'].unique().tolist()
        bounds = metrics.groupby('hasher_name')[[
            'distance_to_closest_image', 'distance_to_closest_incorrect_image'
        ]].max().max(axis=1)
        if grouping:
            group_names = [
                ':'.join(map(str, row.values))
                for idx, row in metrics[grouping].drop_duplicates().iterrows()
            ]
        else:
            group_names = ['']
        ncols = len(hasher_names)
        nrows = len(group_names)

        fig, axs = plt.subplots(
            ncols=ncols,
            nrows=nrows,
            figsize=(ncols * 4, nrows * 3),
            sharey=True)

        for group_name, subset in metrics.groupby(['hasher_name'] + grouping):
            # Get names of group and hasher
            if grouping:
                hasher_name = group_name[0]
                group_name = ':'.join(map(str, group_name[1:]))
            else:
                hasher_name = group_name
                group_name = ''

            # Get the correct axis.
            colIdx = hasher_names.index(hasher_name)
            rowIdx = group_names.index(group_name)
            if ncols > 1 and nrows > 1:
                ax = axs[rowIdx, colIdx]
            elif ncols == 1 and nrows == 1:
                ax = axs
            else:
                ax = axs[rowIdx if nrows > 1 else colIdx]

            # Plot the charts
            pos, neg = subset.groupby(['guid', 'transform_name'])[[
                'distance_to_closest_correct_image',
                'distance_to_closest_incorrect_image'
            ]].min().values.T
            optimal_threshold, _, optimal_recall = compute_threshold_fpr_recall(
                pos=pos, neg=neg, fpr_threshold=fpr_threshold)
            optimal_threshold = optimal_threshold.round(3)
            emd = stats.wasserstein_distance(pos, neg).round(2)
            ax.hist(neg, label='neg', bins=10)
            ax.hist(pos, label='pos', bins=10)
            ax.text(
                0.5,
                0.5,
                f'Recall: {optimal_recall:.0f}% @ {optimal_threshold}\nemd: {emd:.2f}',
                horizontalalignment='center',
                color='black',
                verticalalignment='center',
                transform=ax.transAxes,
                fontsize=12,
                fontweight=1000)
            ax.set_xlim(-0.05 * bounds[hasher_name], bounds[hasher_name])
            if rowIdx == 0:
                ax.set_title(hasher_name)
                ax.legend()
            if colIdx == 0:
                ax.set_ylabel(group_name)
        fig.tight_layout()

    def compute_threshold_recall(self,
                                 fpr_threshold=0.001,
                                 grouping=None,
                                 **kwargs) -> pd.DataFrame:
        """Compute a table for threshold and recall for each category, hasher,
        and transformation combinations. Additional arguments passed to compute_metrics.

        Args:
            fpr_threshold: The false positive rate threshold to use
                for choosing a distance threshold for each hasher.
            grouping: List of fields to group by. By default, all fields are used
                (category, and transform_name).

        Returns:
            A pandas DataFrame with 7 columns. The key columns are threshold
            (The optimal distance threshold for detecting a match for this
            combination), recall (the number of correct matches divided by
            the number of possible matches), and precision (the number correct
            matches divided by the total number of matches whether correct
            or incorrect).
        """
        if grouping is None:
            grouping = ['category', 'transform_name']

        def group_func(subset):
            pos, neg = subset.groupby(['guid', 'transform_name'])[[
                'distance_to_closest_correct_image',
                'distance_to_closest_incorrect_image'
            ]].min().values.T
            optimal_threshold, optimal_fpr, optimal_recall = compute_threshold_fpr_recall(
                pos=pos, neg=neg, fpr_threshold=fpr_threshold)
            return pd.Series({
                'threshold': optimal_threshold,
                'recall': optimal_recall,
                'fpr': optimal_fpr,
                'n_exemplars': len(subset)
            })

        return self.compute_metrics(
            **kwargs).groupby(grouping + ['hasher_name']).apply(group_func)


class BenchmarkDataset(Saveable):
    """A dataset of images separated into
    categories. It is essentially a wrapper around a pandas
    dataframe with the following columns:

    - filepath
    - category
    """

    expected_columns = ['filepath', 'category']

    @classmethod
    def from_tuples(cls, files: typing.List[typing.Tuple[str, str]]):
        """Build dataset from a set of files.

        Args:
            files: A list of tuples where each entry is a pair
                filepath and category.
        """
        df = pd.DataFrame.from_records([{
            'filepath': f,
            'category': c
        } for f, c in files])
        return cls(df)

    def transform(self, transforms, storage_dir, errors):
        raise NotImplementedError()


class BenchmarkTransforms(Saveable):
    """A dataset of transformed images. Essentially wraps a DataFrame with the
    following columns:

    - guid
    - filepath
    - category
    - transform_name
    - input_filepath (for memo purposes only)
    """

    expected_columns = [
        'filepath', 'category', 'transform_name', 'input_filepath', 'guid'
    ]

    def compute_hashes(self, hashers, max_workers):
        raise NotImplementedError()



# pylint: disable=invalid-name
import logging

import numpy as np
import cv2

from .. import tools
from ...utils import flatten
from ..hasher import VideoHasher
from ..image.phash import PHashU8
from .tmk import TMKL1

logger = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes
class SimpleSceneDetection(VideoHasher):
    """The SimpleSceneDetection hasher is a wrapper around other video hashers
    to create separate hashes for different scenes / shots in a video. It works
    by shrinking each frame, blurring it, and doing a simple delta with the previous
    frame. If they are different, this marks the start of a new scene. In addition,
    this wrapper will also remove letterboxing from videos by checking for solid
    black areas on the edges of the frame.

    Args:
        base_hasher: The base video hasher to use for each scene.
        interscene_threshold: The distance threshold between sequential scenes that
            new hashes must meet to be included (this is essentially for deduplication)
        min_frame_size: The minimum frame size to use for computing hashes. This is
            relevant for letterbox detection as black frames will tend to be completely
            "cropped" and make the frame very small.
        max_scene_length: The maximum length of a single scene.
        similarity_threshold: The threshold for detecting whether two frames are
            different enough to constitute a new scene.
    """
    returns_multiple = True

    # pylint: disable=too-many-arguments
    def __init__(self,
                 base_hasher: VideoHasher = None,
                 interscene_threshold=None,
                 min_frame_size=50,
                 similarity_threshold=0.95,
                 max_scene_length=None):
        if base_hasher is None:
            base_hasher = TMKL1(
                frames_per_second=2,
                frame_hasher=PHashU8(
                    exclude_first_term=False, freq_shift=1, hash_size=12),
                distance_metric='euclidean',
                dtype='uint8',
                norm=None,
                quality_threshold=90)
            if interscene_threshold is None:
                interscene_threshold = 50
        if interscene_threshold is not None and base_hasher.returns_multiple:
            raise ValueError(
                'Interscene thresholds not supported for hashers returning multiple hashes.'
            )
        self.base_hasher = base_hasher
        self.frames_per_second = base_hasher.frames_per_second
        self.distance_metric = base_hasher.distance_metric
        self.dtype = base_hasher.dtype
        self.hash_length = base_hasher.hash_length
        self.max_scene_length = max_scene_length
        self.interscene_threshold = interscene_threshold
        self.min_frame_size = min_frame_size
        self.similarity_threshold = similarity_threshold

    def compute_batches(self,
                        filepath,
                        errors='raise',
                        hash_format='base64',
                        batch_size=10):
        """Compute a hash for a video at a given filepath and
        yield hashes in a given batch size.

        Args:
            filepath: Path to video file
            errors: One of "raise", "ignore", or "warn". Passed
                to perception.hashers.tools.read_video.
            hash_format: The hash format to use when returning hashes.
            batch_size: The minimum number of hashes to include in each batch.
        """

        def convert(scenes):
            if hash_format == 'vector':
                return scenes
            if self.base_hasher.returns_multiple:
                return [([
                    self.vector_to_string(h, hash_format=hash_format)
                    for h in hs
                ], frames) for hs, frames in scenes]
            return [(self.vector_to_string(h, hash_format=hash_format), frames)
                    for h, frames in scenes]

        state = None
        for frame, frame_index, frame_timestamp in tools.read_video(
                filepath=filepath,
                frames_per_second=self.frames_per_second,
                errors=errors):
            state = self.process_frame(
                frame=frame,
                frame_index=frame_index,
                frame_timestamp=frame_timestamp,
                state=state,
                batch_mode=True)
            if len(state['scenes']) >= batch_size:
                yield convert(state['scenes'])
                state['scenes'] = []
        assert state is not None
        if state['substate']:
            self.handle_scene(state)
        if state['scenes']:
            yield convert(state['scenes'])

    def handle_scene(self, state, frame_timestamp=None):
        subhash = self.base_hasher.hash_from_final_state(state['substate'])
        if subhash is not None and (self.base_hasher.returns_multiple or (
            (self.interscene_threshold is None or not state['scenes']
             or self.compute_distance(state['scenes'][-1][0],
                                      subhash) > self.interscene_threshold))):
            state['scenes'].append((subhash, state['frames']))
        state['substate'] = None
        state['bounds'] = None
        state['frames'] = []
        state['pre'] = None
        if frame_timestamp is not None:
            state['start'] = frame_timestamp

    def crop(self, frame, bounds):
        # Check to see we have set bounds for this scene yet.
        if not bounds:
            # We don't have bounds, so we'll set them.
            bounds = tools.unletterbox(frame)
            # If the bounds come back invalid (i.e., the frame is too small)
            # or no bounds are found (i.e., the frame is all back), we
            # return None.
            if bounds is None or min(bounds[0][1] - bounds[0][0], bounds[1][1]
                                     - bounds[1][0]) < self.min_frame_size:
                return None, None, None
        (x1, x2), (y1, y2) = bounds
        cropped = np.ascontiguousarray(frame[y1:y2, x1:x2])
        current = cv2.resize(
            cv2.cvtColor(cropped, cv2.COLOR_RGB2GRAY), (128, 128))
        current = cv2.blur(current, ksize=(4, 4))
        return cropped, current, bounds

    # pylint: disable=arguments-differ,too-many-arguments
    def process_frame(self,
                      frame,
                      frame_index,
                      frame_timestamp,
                      state=None,
                      batch_mode=False):
        if not state:
            state = {
                'pre': None,
                'substate': None,
                'start': 0,
                'bounds': None,
                'frames': [],
                'scenes': []
            }
        cropped, current, state['bounds'] = self.crop(frame, state['bounds'])
        if cropped is None:
            # A good crop was not found so we set the start of the scene to this
            # point and continue on to the next frame. This will repeat until we
            # find appropriate bounds.
            state['start'] = frame_timestamp
            return state

        # Check if we have a previous frame to compare the
        # current frame to.
        if state['pre'] is not None:
            # Compute similarity between the previous frame and the
            # current frame.
            similarity = 1 - np.abs(state['pre'].astype('float32') - current.
                                    astype('float32')).sum() / (255 * 128**2)
            if (similarity < self.similarity_threshold or
                (self.max_scene_length is not None and
                 frame_timestamp - state['start'] > self.max_scene_length)):
                # The similarity is too low. We've started a new scene.
                self.handle_scene(state, frame_timestamp)
                cropped, current, state['bounds'] = self.crop(
                    frame, state['bounds'])
                if cropped is None:
                    # See comment above about invalid crops.
                    state['start'] = frame_timestamp
                    return state

        state['pre'] = current
        try:
            state['substate'] = self.base_hasher.process_frame(
                cropped, frame_index, frame_timestamp, state=state['substate'])
            if batch_mode:
                state['frames'].append((frame, frame_index, frame_timestamp))
        except Exception as e:  # pylint: disable=broad-except
            logger.warning('An error occurred while processing a frame: %s',
                           str(e))
        return state

    def hash_from_final_state(self, state):
        if state['substate']:
            self.handle_scene(state)
        if not self.base_hasher.returns_multiple:
            return [h for h, _ in state['scenes']]
        return flatten([hs for hs, _ in state['scenes']])



# pylint: disable=invalid-name
import pdqhash

from ..hasher import ImageHasher


class PDQHash(ImageHasher):
    """The Facebook PDQ hash. Based on the original implementation located at
    the `official repository <https://github.com/facebook/ThreatExchange>`_.
    """
    distance_metric = 'hamming'
    dtype = 'bool'
    hash_length = 256

    def _compute(self, image):
        return pdqhash.compute(image)[0] > 0

    def _compute_with_quality(self, image):
        hash_vector, quality = pdqhash.compute(image)
        return hash_vector > 0, quality

    # pylint: disable=no-self-use
    def _compute_isometric(self, image):
        hash_vectors, _ = pdqhash.compute_dihedral(image)
        names = ['r0', 'r90', 'r180', 'r270', 'fv', 'fh', 'r90fv', 'r90fh']
        return dict(zip(names, hash_vectors))


class PDQHashF(PDQHash):
    dtype = 'float32'
    distance_metric = 'euclidean'
    hash_length = 256

    # pylint: disable=no-self-use
    def _compute(self, image):
        return pdqhash.compute_float(image)



# distutils: extra_compile_args=-fopenmp
# distutils: extra_link_args=-fopenmp
# cython: language_level=3

import math
import numpy as np
import cython
from cython.parallel import prange, parallel
from libc.stdlib cimport abort, malloc, free

cimport numpy as np
cdef extern from "limits.h":
    int INT_MAX

ctypedef np.uint8_t uint8

@cython.boundscheck(False)
@cython.wraparound(False)
def compute_euclidean_pairwise_duplicates(int[:, :] X, float threshold, counts: int[:] = None, compute_overlap=False):
    """Find the pairwise overlap within an array of vectors, where there may be multiple
    vectors for the same file. This function is faster than using scipy.spatial.distance
    because it computes distances in parallel, avoids computing full distances when they're
    not necessary, skips computing distances for pairs of hashes that are for the
    same file, and skips computing distances for vectors if both have already been matched.
    
    Args:
        X: The vectors with shape (N, D). Vectors for the same file need to be
            supplied sequentially so that we can use the counts argument
            to determine which vectors are for the same file.
        counts: For each file, the number of sequential vectors in X. If not
            provided, each vector is assumed to be for a different file (i.e.,
            this is equivalent to `counts = np.ones(N)`).
        compute_overlap: If True, the values returned will be divided by the number
            of hashes in each file. If False, the raw duplicate counts will
            be returned.
    
    Returns:
        duplicates: An array of shape (M!/(2*((M-2)!)), 2) indicating
            the fraction of vectors for each file found in another file.
            The indexing matches that of scipy.spatial.pdist. M is the number of files.
            So if M = 4, the array will represent comparisons of the file indexes as follows:
            [(0, 1), (0, 2), (0, 3), (1, 2), (1, 3)]. So (assuming compute_overlap=True),
            a possible return would be [(1.0, 1.0), (0, 0), (0, 0), (0.66, 1.0), (0.5, 0.25)]
            which means that:

            - There was 100% overlap between file 0 and file 1
            - 66% of file 1 was in file 2 and 100% of file 2 was in file 1
            - 50% of file 2 was in file 3 and 25% of file 3 was in file 2
    """
    if counts is None:
        counts_arr = np.ones(X.shape[0], dtype=np.int32)
        counts = counts_arr
    cdef Py_ssize_t n = X.shape[0]
    cdef Py_ssize_t m = counts.shape[0]
    cdef Py_ssize_t d = X.shape[1]
    cdef Py_ssize_t n_pairs = int(math.factorial(m)/(2*math.factorial(m-2)))
    cdef Py_ssize_t max_counts = np.max(counts)
    cdef int compute_overlap_int = 0
    if compute_overlap:
        compute_overlap_int = 1
    # i_1 is the index of file1, i_2 is the index of file2, i_d is the
    # index of the vector dimension we're on, i_i is used to compute
    # the starting index in the flattened vector in the different threads.
    # i_1_subhash is the index of the hash on file1, i_2_subhash is
    # the index of the hash on file2.
    cdef Py_ssize_t i_1, i_2, i_d, i_i, i_1_sub, i_2_sub, i_1_offset
    duplicate_arr = np.zeros((n_pairs, 2), dtype=np.double)
    cdef double[:, :] duplicate = duplicate_arr
    offsets_arr = np.zeros(m, dtype=np.int32)
    cdef np.int32_t[:] offsets = offsets_arr
    for i_1 in range(m):
        for i_i in range(i_1):
            offsets[i_1] += counts[i_i]
    cdef size_t local_buf_size = 4 # distance, flattened array offset, index_offset_1, index_offset_2
    cdef float threshold2 = threshold ** 2
    with nogil, parallel():
        local_buf = <int *> malloc(sizeof(int) * local_buf_size)

        # An array of flags indicating whether a vector in file 1 was
        # matched.
        matched_1 = <int *> malloc(sizeof(int) * max_counts)

        # An array of flags indicating whether a vector in file 2 was
        # matched.
        matched_2 = <int *> malloc(sizeof(int) * max_counts)
        if local_buf is NULL or matched_1 is NULL or matched_2 is NULL:
            abort()
        # Iterate over all of the files.
        for i_1 in prange(m-1):
            local_buf[1] = 0
            local_buf[2] = offsets[i_1]
            # Compute the index of the output vector
            # where we will count the number of duplicates.
            for i_i in range(i_1):
                local_buf[1] += m - i_i - 1
            # Iterate over all the other files to compare.
            for i_2 in range(i_1 + 1, m):
                local_buf[3] = offsets[i_2]
                # Initialize all match flags to zero for
                # both file 1 and file 2.
                for i_1_sub in range(counts[i_1]):
                    matched_1[i_1_sub] = 0
                for i_2_sub in range(counts[i_2]):
                    matched_2[i_2_sub] = 0
                # Iterate over all the hashes in file1
                for i_1_sub in range(counts[i_1]):
                    # Iterate over all the hashes in file2
                    for i_2_sub in range(counts[i_2]):
                        local_buf[0] = 0
                        if matched_1[i_1_sub] == 1 and matched_2[i_2_sub] == 1:
                            # Both the vectors in this pair have already been matched, so
                            # there is nothing to gain from this comparison.
                            continue
                        for i_d in range(d):
                            local_buf[0] += (X[local_buf[2] + i_1_sub, i_d] - X[local_buf[3] + i_2_sub, i_d]) ** 2
                            if local_buf[0] > threshold2:
                                # If we're already beyond the distance threshold,
                                # we don't need to continue computing squared
                                # distances.
                                break
                        if local_buf[0] < threshold2:
                            # A match was found. Set flags for both vectors
                            # to 1.
                            matched_1[i_1_sub] = 1
                            matched_2[i_2_sub] = 1
                # Add up the number of matches for file 1.
                for i_1_sub in range(counts[i_1]):
                    duplicate[local_buf[1], 0] += matched_1[i_1_sub]
                # Add up the number of matches for file 2.
                for i_2_sub in range(counts[i_2]):
                    duplicate[local_buf[1], 1] += matched_2[i_2_sub]
                # Divide by the total number of vectors for each file.
                if compute_overlap_int:
                    duplicate[local_buf[1], 0] /= counts[i_1]
                    duplicate[local_buf[1], 1] /= counts[i_2]
                # Advance to the next pair index.
                local_buf[1] += 1
        free(local_buf)
        free(matched_1)
        free(matched_2)
    return duplicate_arr



import cv2
import numpy as np

import pywt

from ..hasher import ImageHasher


class WaveletHash(ImageHasher):
    """Similar to PHash but using wavelets instead of DCT.
    Implementation based on that of
    `ImageHash <https://github.com/JohannesBuchner/imagehash>`_.
    """
    distance_metric = 'hamming'
    dtype = 'bool'

    def __init__(self, hash_size=8, image_scale=None, mode='haar'):
        assert hash_size & (
            hash_size - 1) == 0, "Hash size must be a power of 2."
        if image_scale is not None:
            assert (
                image_scale &
                (image_scale - 1) == 0), "Image scale must be a power of 2."
            assert image_scale >= hash_size, \
                "Image scale must be greater than or equal to than hash size."
        self.hash_size = hash_size
        self.image_scale = image_scale
        self.mode = mode
        self.hash_length = hash_size * hash_size

    def _compute(self, image):
        if self.image_scale is None:
            image_scale = max(2**int(np.log2(min(image.shape[:2]))),
                              self.hash_size)
        else:
            image_scale = self.image_scale
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        image = cv2.resize(
            image,
            dsize=(image_scale, image_scale),
            interpolation=cv2.INTER_AREA)
        image = np.float32(image) / 255

        ll_max_level = int(np.log2(image_scale))
        level = int(np.log2(self.hash_size))
        dwt_level = ll_max_level - level

        if self.mode == 'haar':
            coeffs = pywt.wavedec2(image, 'haar', level=ll_max_level)
            coeffs = list(coeffs)
            coeffs[0] *= 0
            image = pywt.waverec2(coeffs, 'haar')

        coeffs = pywt.wavedec2(image, self.mode, level=dwt_level)
        dwt_low = coeffs[0]

        # Subtract median and compute hash
        med = np.median(dwt_low)
        diff = dwt_low > med

        return diff.flatten()



# pylint: disable=too-many-arguments,too-many-locals

import os
import json
import base64
import typing
import warnings
import urllib.parse
import urllib.request

try:
    import tqdm  # pylint: disable=unused-import
except ImportError:  # pragma: no cover
    tqdm = None
from scipy import spatial
import numpy as np

from . import hashers as perception_hashers
from .utils import flatten
try:
    from . import extensions  # type: ignore
except ImportError:
    warnings.warn(
        'C extensions were not built. Some metrics will be computed more slowly. '
        'Please install from wheels or set up a compiler prior to installation '
        'from source to use extensions.')
    extensions = None


def _multiple_hashes_for_ids(
        hashes: typing.List[typing.Tuple[str, typing.Union[str, np.ndarray]]]):
    """Check if a list of (hash_id, hash) tuples has more
    than one hash for a hash_id.

    Args:
        hashes: A list of (hash_id, hash) tuples.
    """
    hash_ids = [hash_id for hash_id, _ in hashes]
    return len(hash_ids) != len(set(hash_ids))


# pylint: disable=too-many-branches,too-many-statements
def deduplicate_hashes(
        hashes: typing.List[typing.Tuple[str, typing.Union[str, np.ndarray]]],
        threshold: float,
        hash_format: str = 'base64',
        hasher: perception_hashers.ImageHasher = None,
        hash_length: int = None,
        hash_dtype: str = None,
        distance_metric: str = None,
        progress: 'tqdm.tqdm' = None) -> typing.List[typing.Tuple[str, str]]:
    """Find duplicates using a list of precomputed hashes.

    Args:
        hashes: A list of (id, hash) tuples
        threshold: A distance threshold
        hasher: A hasher to use for computing distances
        progress: A tqdm object for reporting progress

    Returns:
        A list of duplicated id pairs. To use, you can just remove the
        first entry of each pair from your dataset. The pairs are provided
        in the event that you wish to apply further analysis.
    """
    assert (
        hash_length is not None and
        hash_dtype is not None and
        distance_metric is not None
    ) or (
        hasher is not None
    ), \
        (
            'You must provide either `hasher` or all of '
            '`hash_length`, `hash_dtype`, and `distance_metric`.'
        )
    if hasher is not None:
        assert all(k is None for k in [hash_length, hash_dtype, distance_metric]), \
            'If hasher is provided, hash_length, hash_dtype, and distance_metric must all be None.'
        hash_length = hasher.hash_length
        hash_dtype = hasher.dtype
        distance_metric = hasher.distance_metric
    assert hash_length is not None
    assert isinstance(hash_dtype, str)
    assert isinstance(distance_metric, str)
    # If there is more than one hash for an id, we want them
    # to be sequential in case we are able to use the more
    # efficient distance calculation (compute_euclidean_pairwise_duplicates)
    # that skips computation of distance between two hashes for the same file.
    multiple_hashes_per_id = _multiple_hashes_for_ids(hashes)
    if multiple_hashes_per_id:
        hashes = sorted(hashes)
    vectors = np.array([
        perception_hashers.tools.string_to_vector(
            hash_string=hash_string_or_vector,
            hash_format=hash_format,
            hash_length=hash_length,
            dtype=hash_dtype)
        if isinstance(hash_string_or_vector, str) else hash_string_or_vector
        for _, hash_string_or_vector in hashes
    ])
    files = np.array([identifier for identifier, _ in hashes])
    pairs: typing.List[typing.Tuple[str, str]] = []
    n_hashes = len(vectors)
    start_idx = 0
    end_idx = None
    if distance_metric != 'euclidean' or 'int' not in hash_dtype or extensions is None:
        iterator = range(n_hashes)
        if progress is not None:
            iterator = progress(
                iterator, total=n_hashes, desc='Deduplicating.')
        distances = spatial.distance.pdist(vectors, metric=distance_metric)
        for hash_index in iterator:
            if end_idx is not None:
                start_idx = end_idx
            end_idx = start_idx + (n_hashes - hash_index - 1)
            current_distances = distances[start_idx:end_idx]
            duplicated_files = files[hash_index +
                                     1:][current_distances < threshold]
            current_file = files[hash_index]
            # We have to make sure the two files are not the same file
            # because it can happen for highly symmetric images when
            # we are including isometric hashes.
            pairs.extend([(current_file, duplicated_file)
                          for duplicated_file in duplicated_files
                          if duplicated_file != current_file])
    else:
        # We want to count the number of hashes for each unique hash ID. There
        # may be more than one -- for example in the case of video. We need
        # this so we can pass it to the compute_euclidean_pairwise_duplicates
        # function.
        if multiple_hashes_per_id:
            counts = np.zeros(
                shape=len(set(
                    hash_id for hash_id, _ in hashes))).astype('int32')
            previous_hash_id = None
            counts_idx = 0
            files = [
            ]  # We're going to re-build the IDs with deduplicated files.
            for hash_id, _ in hashes:
                if hash_id != previous_hash_id:
                    files.append(hash_id)
                if previous_hash_id is not None and hash_id != previous_hash_id:
                    counts_idx += 1
                counts[counts_idx] += 1
                previous_hash_id = hash_id
            files = np.array(files)
        else:
            counts = None
        n_files = len(files)
        iterator = range(n_files)
        if progress is not None:
            iterator = progress(iterator, total=n_files, desc='Deduplicating.')
        duplicated = (extensions.compute_euclidean_pairwise_duplicates(
            vectors.astype('int32'), threshold=threshold,
            counts=counts).max(axis=1) > 0)
        for file_index in iterator:
            if end_idx is not None:
                start_idx = end_idx
            end_idx = start_idx + (n_files - file_index - 1)
            current_duplicated = duplicated[start_idx:end_idx]
            current_file = files[file_index]
            duplicated_files = files[file_index + 1:][current_duplicated]
            pairs.extend([(current_file, duplicated_file)
                          for duplicated_file in duplicated_files
                          if duplicated_file != current_file])
    return list(set(pairs))


# pylint: disable=too-many-locals
def deduplicate(
        files: typing.List[str],
        hashers: typing.List[typing.Tuple[perception_hashers.
                                          ImageHasher, float]],
        isometric: bool = False,
        progress: 'tqdm.tqdm' = None) -> typing.List[typing.Tuple[str, str]]:
    """Find duplicates in a list of files.

    Args:
        files: A list of filepaths.
        hashers: A list of tuples of the form (hasher, threshold)
        isometric: Whether to compare the rotated versions of the images
        progress: A tqdm progress indicator

    Returns:
        A list of duplicated file pairs. To use, you can just remove the
        first entry of each pair from your dataset. The pairs are provided
        in the event that you wish to apply further analysis.
    """
    files_dedup = set(files)
    if len(files_dedup) != len(files):
        warnings.warn(
            message=
            "Duplicate file paths were provided. These will be automatically removed.",
            category=UserWarning)
        files = list(files_dedup)
    pairs: typing.List[typing.Tuple[str, str]] = []
    for hasher_idx, (hasher, threshold) in enumerate(hashers):
        hash_dicts = hasher.compute_parallel(
            filepaths=files,
            progress=progress,
            progress_desc=
            f"Computing hashes for hash {hasher_idx+1} of {len(hashers)}.",
            isometric=isometric)
        hash_list = sorted(hash_dicts, key=lambda h: h['filepath'])
        if isometric:
            hash_list = flatten([
                list(row['hash'].values()) for row in hash_dicts
                if row['error'] is None
            ])
            files_for_hashes = flatten([[row['filepath']] * 8
                                        for row in hash_dicts
                                        if row['error'] is None])
        elif hasher.returns_multiple:
            hash_list = flatten(
                [row['hash'] for row in hash_dicts if row['error'] is None])
            files_for_hashes = flatten([[row['filepath']] * 8
                                        for row in hash_dicts
                                        if row['error'] is None])
        else:
            hash_list = [
                row['hash'] for row in hash_dicts if row['error'] is None
            ]
            files_for_hashes = [
                row['filepath'] for row in hash_dicts if row['error'] is None
            ]
        pairs.extend(
            deduplicate_hashes(
                hashes=list(zip(files_for_hashes, hash_list)),
                hasher=hasher,
                threshold=threshold,
                progress=progress))
    return list(set(pairs))


# pylint: disable=too-few-public-methods
class SaferMatcher:
    """An object for matching hashes with the known CSAM hashes in the
    Safer matching service.
    Please contact `info@getsafer.io <mailto:info@getsafer.io>`_
    for details on obtaining credentials and information on how match
    responses are provided.

    Here's a minimalist example:

    .. code-block:: python

        from perception import hashers, tools

        hasher = hashers.PHash(hash_size=16)
        matches = hashers.tools.SaferMatcher(
            api_key='YOUR_API_KEY',
            username='YOUR_USERNAME', # You only need to provide
            password='YOUR_PASSWORD', # an API key OR username/password.
            url='MATCHING_SERVICE_URL'
        )

    For authentication, you must provide the API key OR username and password pair.
    If neither is provided, the function will attempt to find them as environment
    variables with names :code:`SAFER_MATCHING_SERVICE_API_KEY`,
    :code:`SAFER_MATCHING_SERVICE_USERNAME`, and :code:`SAFER_MATCHING_SERVICE_PASSWORD`,
    respectively. You must also provide the URL endpoint for the matching service,
    either as a keyword argument or as a :code:`SAFER_MATCHING_SERVICE_URL`
    environment variable.

    Args:
        api_key: A base64 encoded set of matching service credentials
        username: Matching service username
        password: Matching service password
        url: Safer matching service URL
        hasher: A hasher to use for matching
        hasher_api_id: The hasher ID for finding matches.
        quality_threshold: The quality threshold filter to use
    """

    def __init__(self,
                 api_key: str = None,
                 username: str = None,
                 password: str = None,
                 url: str = None,
                 hasher: perception_hashers.ImageHasher = None,
                 hasher_api_id: str = None,
                 quality_threshold: int = 90):
        if (username is None and password is None and api_key is None and
                os.environ.get('SAFER_MATCHING_SERVICE_USERNAME') is not None
                and
                os.environ.get('SAFER_MATCHING_SERVICE_PASSWORD') is not None):
            username = os.environ['SAFER_MATCHING_SERVICE_USERNAME']
            password = os.environ['SAFER_MATCHING_SERVICE_PASSWORD']
        if username is not None and password is not None:
            credentials = ('%s:%s' % (username, password))
            api_key = base64.b64encode(
                credentials.encode('ascii')).decode('ascii')
        if api_key is None:
            api_key = os.environ.get('SAFER_MATCHING_SERVICE_API_KEY')
            if api_key is None:
                raise ValueError(
                    'You must provide one of (1) API key, (2) API key provided as '
                    '`SAFER_MATCHING_SERVICE_API_KEY` env var, (3) username and password or '
                    '(4) username and password as `SAFER_MATCHING_SERVICE_USERNAME` and '
                    '`SAFER_MATCHING_SERVICE_PASSWORD` env vars.')
        if url is None:
            url = os.environ.get('SAFER_MATCHING_SERVICE_URL')
            if url is None:
                raise ValueError(
                    'You must provide either the url or the SAFER_MATCHING_SERVICE_URL env var.'
                )
        if urllib.parse.urlparse(url).scheme != 'https' and not os.environ.get(
                'SAFER_MATCHING_SERVICE_DEV_ALLOW_HTTP'):
            raise ValueError(
                'You must provide an url that begins with `https://`.')
        self.api_key = api_key
        self.url = url
        if hasher is None:
            hasher = perception_hashers.PHash(hash_size=16, highfreq_factor=4)
        if hasher_api_id is None:
            hasher_api_id = 'phash'
        self.hasher = hasher
        self.hasher_api_id = hasher_api_id
        self.quality_threshold = quality_threshold

    def match(self, images: typing.List[
            typing.Union[str, typing.Tuple[perception_hashers.tools.
                                           ImageInputType, str]]]) -> dict:
        """Match hashes with the Safer matching service.

        Args:
            images: A list of image filepaths or (image_like, image_id) tuples.

        Returns:
            A dictionary of matches. See Safer matching service documentation (
            contact Thorn for a copy).
        """
        raw_hashes = [
            self.hasher.compute_with_quality(
                image if isinstance(image, str) else image[0])
            for image in images
        ]
        hashes = [{
            'id':
            image if isinstance(image, str) else image[1],
            self.hasher_api_id:
            hash_string,
            'md5':
            perception_hashers.tools.compute_md5(image) if isinstance(
                image, str) else
            (perception_hashers.tools.compute_md5(image[0]) if isinstance(
                image[0], str) else None)
        } for image, (hash_string, quality) in zip(images, raw_hashes)
                  if quality > self.quality_threshold]
        for hash_dict in hashes:
            # We cannot include an md5 key if we don't
            # have the md5.
            if hash_dict['md5'] is None:
                del hash_dict['md5']
        if not hashes:
            warnings.warn(
                message="No images of sufficient quality were found.",
                category=UserWarning)
            return {}
        body = {'hashes': hashes, 'version': 'v2'}
        headers = {
            'Authorization': f'Basic {self.api_key}',
            'Content-Type': 'application/json'
        }
        req = urllib.request.Request(
            url=self.url,
            data=str(json.dumps(body)).encode('utf-8'),
            headers=headers,
            method='POST')
        res = urllib.request.urlopen(req)
        return json.loads(res.read().decode('utf-8'))



# -*- coding: utf-8 -*-
import warnings
from distutils.command.build_ext import build_ext
from distutils.errors import CCompilerError, DistutilsExecError, DistutilsPlatformError
from setuptools import setup
from Cython.Build import cythonize
import numpy

import versioneer

class BuildFailure(Exception):
    pass

class CatchableBuildExt(build_ext):
    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError:
            raise BuildFailure()

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except (CCompilerError, DistutilsExecError, DistutilsPlatformError):
            raise BuildFailure()

try:
    setup(version=versioneer.get_version(),
          cmdclass={**versioneer.get_cmdclass(), 'build_ext': CatchableBuildExt},
          ext_modules=cythonize(
              "perception/**/extensions.pyx",
          ), include_dirs=[numpy.get_include()])
except BuildFailure:
    warnings.warn('Failed to build Cython extensions. They will not be available at runtime.')
    setup(version=versioneer.get_version(),
          cmdclass=versioneer.get_cmdclass())



# pylint: disable=too-many-arguments,too-many-branches

import os
import typing

import cv2
import ffmpeg

from ..hashers.tools import read_video


def probe(filepath):
    """Get the output of ffprobe."""
    return ffmpeg.probe(filepath)


def sanitize_output_filepath(input_filepath, output_filepath, output_ext=None):
    """Get a suitable output filepath with an extension based on
    an input filepath.

    Args:
        input_filepath: The filepath for the source file.
        output_filepath: The filepath for the output file.
        output_ext: A new extension to add (e.g., '.gif')
    """
    _, input_ext = os.path.splitext(input_filepath)
    if not output_filepath.lower().endswith(output_ext or input_ext):
        output_filepath += output_ext or input_ext
    return output_filepath


def get_simple_transform(width: typing.Union[str, int] = -1,
                         height: typing.Union[str, int] = -1,
                         pad: str = None,
                         codec: str = None,
                         clip_pct: typing.Tuple[float, float] = None,
                         clip_s: typing.Tuple[float, float] = None,
                         sar=None,
                         fps=None,
                         output_ext=None):
    """Resize to a specific size and re-encode.

    Args:
        width: The target width (-1 to maintain aspect ratio)
        height: The target height (-1 to maintain aspect ratio)
        pad: An ffmpeg pad argument provided as a string.
        codec: The codec for encoding the video.
        fps: The new frame rate for the video.
        clip_pct: The video start and end in percentages of video duration.
        clip_s: The video start and end in seconds (used over clip_pct if both
            are provided).
        sar: Whether to make all videos have a common sample aspect
            ratio (i.e., for all square pixels, set this to '1/1').
        output_ext: The extension to use when re-encoding (used to select
            video format). It should include the leading '.'.
    """

    def transform(input_filepath, output_filepath):
        output_filepath = sanitize_output_filepath(input_filepath,
                                                   output_filepath, output_ext)
        data = None
        if codec is None:
            data = (data or probe(input_filepath))
            output_codec = [
                s for s in data['streams'] if s['codec_type'] == 'video'
            ][0]['codec_name']
        else:
            output_codec = codec
        format_kwargs = {'codec:v': output_codec}
        if clip_pct is not None or clip_s is not None:
            pct_start, pct_end, pos_start, pos_end = None, None, None, None
            if clip_pct is not None:
                pct_start, pct_end = clip_pct
            if clip_s is not None:
                pos_start, pos_end = clip_s
            if pct_start is not None:
                assert 0 <= pct_start <= 1, 'Start position must be between 0 and 1.'
            if pct_end is not None:
                assert 0 <= pct_end <= 1, 'End position must be between 0 and 1.'
            if pct_start is not None and pct_end is not None:
                assert pct_start < pct_end, 'End must be greater than start.'
            if (pct_start is not None
                    and pos_start is None) or (pct_end is not None
                                               and pos_end is None):
                # We only want to get the duration for the video if we need
                # it.
                data = data or probe(input_filepath)
                duration = float(data['streams'][0]['duration'])
            if pct_start is not None or pos_start is not None:
                format_kwargs[
                    'ss'] = pos_start or pct_start * duration  # type: ignore
            if pct_end is not None or pos_end is not None:
                format_kwargs[
                    't'] = pos_end or pct_end * duration  # type: ignore
        stream = ffmpeg.input(input_filepath)
        if not (width == -1 and height == -1):
            stream = stream.filter('scale', width, height)
        if pad is not None:
            stream = stream.filter('pad', *pad.split(':'))
        if fps is not None:
            stream = stream.filter('fps', fps)
        if sar is not None:
            stream = stream.filter('setsar', sar)
        stream = stream.output(output_filepath, **format_kwargs) \
                    .overwrite_output()
        ffmpeg.run(stream)
        if os.path.isfile(output_filepath):
            return output_filepath
        return None

    return transform


def get_slideshow_transform(frame_input_rate,
                            frame_output_rate,
                            max_frames=None,
                            offset=0):
    """Get a slideshow transform to create slideshows from
    videos.

    Args:
        frame_input_rate: The rate at which frames will be sampled
            from the source video (e.g., a rate of 1 means we collect
            one frame per second of the input video).
        frame_output_rate: The rate at which the sampled frames are played
            in the slideshow (e.g., a rate of 0.5 means each frame will
            appear for 2 seconds).
        max_frames: The maximum number of frames to write.
        offset: The number of seconds to wait before beginning the slide show.
    """

    def transform(input_filepath, output_filepath):
        output_filepath = sanitize_output_filepath(
            input_filepath, output_filepath, output_ext='.mov')
        writer = None
        frame_count = 0
        try:
            for frame, _, timestamp in read_video(
                    filepath=input_filepath,
                    frames_per_second=frame_input_rate):
                if timestamp < offset:
                    continue
                if writer is None:
                    writer = cv2.VideoWriter(
                        filename=output_filepath,
                        fourcc=cv2.VideoWriter_fourcc(*'mjpg'),
                        fps=frame_output_rate,
                        frameSize=tuple(frame.shape[:2][::-1]),
                        isColor=True)
                writer.write(cv2.cvtColor(frame, cv2.COLOR_RGB2BGR))
                frame_count += 1
                if max_frames is not None and frame_count >= max_frames:
                    break
        finally:
            if writer is not None:
                writer.release()
        if os.path.isfile(output_filepath):
            return output_filepath
        return None

    return transform


def get_black_frame_padding_transform(duration_s=0, duration_pct=0):
    """Get a transform that adds black frames at the start and end
    of a video.

    Args:
        duration_s: The duration of the black frames in seconds.
        duration_pct: The duration of the black frames
            as a percentage of video duration. If both duration_s
            and duration_pct are provided, the maximum value
            is used.
    """

    def transform(input_filepath, output_filepath):
        output_filepath = sanitize_output_filepath(input_filepath,
                                                   output_filepath)
        stream = next(
            stream for stream in probe(input_filepath)['streams']
            if stream['codec_type'] == 'video')
        assert stream['sample_aspect_ratio'] == '1:1', 'SAR is not 1:1.'
        width = stream['width']
        height = stream['height']
        duration = max(duration_s, duration_pct * float(stream['duration']))
        ffmpeg.input(input_filepath).output(
            output_filepath,
            vf=("color=c=black:s={width}x{height}:d={duration} [pre] ; "
                "color=c=black:s={width}x{height}:d={duration} [post] ; "
                "[pre] [in] [post] concat=n=3").format(
                    width=width, height=height,
                    duration=duration)).overwrite_output().run()
        if os.path.isfile(output_filepath):
            return output_filepath
        return None

    return transform



from .average import AverageHash
from .phash import PHash, PHashF, PHashU8
from .wavelet import WaveletHash
from .opencv import MarrHildreth, BlockMean, ColorMoment
from .dhash import DHash

__all__ = [
    'AverageHash', 'PHash', 'WaveletHash', 'MarrHildreth', 'BlockMean',
    'ColorMoment', 'DHash', 'PHashF', 'PHashU8'
]



import cv2

from ..hasher import ImageHasher
from .. import tools


class AverageHash(ImageHasher):
    """Computes a simple hash comparing the intensity of each
    pixel in a resized version of the image to the mean.
    Implementation based on that of
    `ImageHash <https://github.com/JohannesBuchner/imagehash>`_."""
    distance_metric = 'hamming'
    dtype = 'bool'

    def __init__(self, hash_size=8):
        assert hash_size >= 2, "Hash size must be greater than or equal to 2."
        self.hash_size = hash_size
        self.hash_length = hash_size * hash_size

    def _compute(self, image):
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        image = cv2.resize(
            image,
            dsize=(self.hash_size, self.hash_size),
            interpolation=cv2.INTER_AREA)
        diff = image > image.mean()
        return diff.flatten()

    def _compute_isometric_from_hash(self, vector):
        return {
            transform_name: diff.flatten()
            for transform_name, diff in tools.get_isometric_transforms(
                vector.reshape(self.hash_size, self.hash_size, 1),
                require_color=False).items()
        }
